"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import os
import sys
import numpy as np
import pandas as pd
from argparse import Namespace
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from pkscreener.classes import Utility, ImageUtility
from PKDevTools.classes.log import default_logger
from pkscreener.classes.ConfigManager import parser, tools
from pkscreener.classes.Portfolio import Portfolio, PortfolioCollection
from pkscreener.classes.PKTask import PKTask
from pkscreener.classes.PKScheduler import PKScheduler
from PKDevTools.classes.OutputControls import OutputControls

configManager = tools()
configManager.getConfig(parser)

def summariseAllStrategies(testing=False):
    reports = getSavedBacktestReportNames(testing=testing)
    df_all = None
    counter = 0
    for report in reports:
        counter += 1
        OutputControls().printOutput(f"Processing {counter} of {len(reports)}...")
        df = bestStrategiesFromSummaryForReport(
            f"PKScreener_{report}_Insights_DateSorted.html", summary=True,includeLargestDatasets=True
        )
        if df is not None:
            df.insert(loc=0, column="Scanner", value=report)
            if df_all is not None:
                df_all = pd.concat([df_all, df], axis=0)
            else:
                df_all = df
        # sys.stdout.write("\x1b[1A")
        OutputControls().moveCursorUpLines(1)
    if df_all is not None:
        df_all = df_all.replace(np.nan, "-", regex=True)
    return df_all

def getSavedBacktestReportNames(testing=False):
    indices = [1,5,8,11,12,14] if not testing else [1]
    scanSkipIndices = [21, 22] if not testing else [1,2,3,4,5,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25]
    indexWithSubindices = [6, 7] if not testing else [6]
    subIndices = {6: [1, 2, 3, 4, 5, 6, 7], 7: [1, 2, 3, 4, 5, 6, 7]} if not testing else {6: [7]}
    indexWithSubLevelindices = {6:[7],7:[6]}
    subLevelIndices = {7: [1, 2, 3], 6: [1, 2, 3]} if not testing else {7: [1]}
    reports = []
    for index in indices:
        scanTypeStartIndex = 1
        scanTypeEndIndex = 25
        reportName = f"B_{index}_"
        while scanTypeStartIndex <= scanTypeEndIndex:
            if scanTypeStartIndex not in scanSkipIndices:
                reportName = f"{reportName}{scanTypeStartIndex}"
                if scanTypeStartIndex in indexWithSubindices:
                    for subIndex in subIndices[scanTypeStartIndex]:
                        subReportName = f"{reportName}_{subIndex}"
                        if subIndex in indexWithSubLevelindices[scanTypeStartIndex]:
                            for subLevelIndex in subLevelIndices[subIndex]:
                                subLevelReportName = f"{subReportName}_{subLevelIndex}"
                                reports.append(subLevelReportName)
                        else:
                            reports.append(subReportName)
                else:
                    reports.append(reportName)
            reportName = f"B_{index}_"
            scanTypeStartIndex += 1
    return reports


def bestStrategiesFromSummaryForReport(reportName: None, summary=False,includeLargestDatasets=False):
    dfs = []
    insights = None
    if "PKDevTools_Default_Log_Level" not in os.environ.keys():
        if (("RUNNER" not in os.environ.keys())):
            return None
    if not configManager.showPastStrategyData:
        return None
    try:
        dfs = pd.read_html(
            "https://pkjmesra.github.io/PKScreener/Backtest-Reports/{0}".format(
                reportName.replace("_X_", "_B_").replace("_G_", "_B_").replace("_S_", "_B_")
            ),encoding="UTF-8", attrs = {'id': 'resultsTable'}
        )
    except Exception as e: # pragma: no cover
        pass
    if dfs is not None and len(dfs) > 0:
        df = dfs[0]
        if len(df) > 0:
            periods = configManager.periodsRange
            insights = cleanupInsightsSummary(df, periods)
            # insights = insights.replace('', np.nan, regex=True)
            # insights = insights.replace('-', np.nan, regex=True)
            dfs = []
            max_best_df = insights.copy()
            max_datasets_df = insights.copy()
            if includeLargestDatasets:
                addLargeDatasetInsights(dfs, max_datasets_df)
            
            insights_list = []
            dfs.append(max_best_df)
            getMaxBestInsight(summary, dfs, periods, insights_list)
            insights = pd.DataFrame(insights_list).drop_duplicates(ignore_index=True)
            insights.dropna(axis=0, how="all", inplace=True)
            insights = insights.replace(np.nan, "-", regex=True)
    return insights

def cleanupInsightsSummary(df, periods):
    df = df.replace(" %", "", regex=True)
    df = df.replace("-", np.nan, regex=True)
    for period in periods:
        df.rename(
                    columns={
                        f"{period}D-%": f"{period}Pd-%",
                        f"{period}D-10k": f"{period}Pd-10k",
                    },
                    inplace=True,
                )
        castToFloat(df, period)
    insights = df[df["ScanType"].astype(str).str.startswith("[SUM]")]
    return insights

def getMaxBestInsight(summary, dfs, periods, insights_list):
    for dfInsights in dfs:
        df = dfInsights.copy()
        df.reset_index(drop = True, inplace=True)
        strategy_percent = {}
        strategy = {}
        firstPeriod = True
        rowIndex = 0
        for prd in periods:
            rowIndex += 1
            try:
                max_p = df[f"{prd}Pd-%"].max()
                maxIndexPos = df[f"{prd}Pd-%"].idxmax()
                bestScanFilter = str(
                            df["ScanType"].iloc[maxIndexPos]).replace("[SUM]", "")
                resultPoints = bestScanFilter.split("(")[-1]
                strategy_percent[f"{prd}-Pd"] = f"{colorText.GREEN if max_p > 0 else (colorText.FAIL if max_p < 0 else colorText.WARN)}{max_p} %{colorText.END}{(' from ('+resultPoints) if (summary or firstPeriod) else ''}"
                scanType = (bestScanFilter.split("(")[0] if not summary else bestScanFilter)
                strategy[f"{prd}-Pd"] = scanType
            except KeyError:
                    max_p = df[rowIndex*2]
                    bestScanFilter = str(df[0]).replace("[SUM]", "")
                    resultPoints = bestScanFilter.split("(")[-1]
                    strategy_percent[f"{prd}-Pd"] = f"{colorText.GREEN if max_p > 0 else (colorText.FAIL if max_p < 0 else colorText.WARN)}{max_p} %{colorText.END}{(' from ('+resultPoints) if (summary or firstPeriod) else ''}"
                    scanType = (bestScanFilter.split("(")[0] if not summary else bestScanFilter)
                    strategy[f"{prd}-Pd"] = scanType
            except Exception:# pragma: no cover
                # default_logger().debug(e, exc_info=True)
                try:
                    max_p = df[f"{prd}Pd-%"]
                    bestScanFilter = str(df["ScanType"]).replace("[SUM]", "")
                    resultPoints = bestScanFilter.split("(")[-1]
                    strategy_percent[f"{prd}-Pd"] = f"{colorText.GREEN if max_p > 0 else (colorText.FAIL if max_p < 0 else colorText.WARN)}{max_p} %{colorText.END}{(' from ('+resultPoints) if (summary or firstPeriod) else ''}"
                    scanType = (bestScanFilter.split("(")[0] if not summary else bestScanFilter)
                    strategy[f"{prd}-Pd"] = scanType
                except Exception as e:# pragma: no cover
                    default_logger().debug(e, exc_info=True)
                    pass
                pass
            if summary:
                strategy[f"{prd}-Pd"] = strategy_percent[f"{prd}-Pd"].split("from")[0] + " " + strategy[f"{prd}-Pd"]
            firstPeriod = False
        insights_list.extend([strategy] if summary else [strategy, strategy_percent])

def addLargeDatasetInsights(dfs, max_datasets_df):
    max_datasets_df[["ScanTypeSplit", "DatasetSize"]] = max_datasets_df[
                    "ScanType"
                ].str.split("(", n=1, expand=True)
    max_datasets_df["DatasetSize"] = max_datasets_df["DatasetSize"].str.replace(")", "")
    try:
        max_datasets_df["DatasetSize"] = (max_datasets_df["DatasetSize"].astype(float).fillna(0.0))
    except Exception as e:# pragma: no cover
        default_logger().debug(e, exc_info=True)
        max_datasets_df.loc[:, "DatasetSize"] = max_datasets_df.loc[:, "DatasetSize"].apply(
                        lambda x: x.split("(")[-1]
                    )
        max_datasets_df["DatasetSize"] = (max_datasets_df["DatasetSize"].astype(float).fillna(0.0))
        pass
    max_size = max_datasets_df["DatasetSize"].max()
    max_datasets_df = max_datasets_df[(max_datasets_df["DatasetSize"] == max_size)].fillna(0.0)
    for i in range(0, len(max_datasets_df)):
        dfs.append(max_datasets_df.iloc[i])

def castToFloat(df, prd):
    if f"{prd}Pd-%" in df.columns:
        df[f"{prd}Pd-%"] = (df[f"{prd}Pd-%"].astype(float).fillna(0.0))


def xRaySummary(savedResults=None):
    if savedResults is None or not isinstance(savedResults, pd.DataFrame) or savedResults.empty:
        return savedResults
    saveResults = savedResults.copy()
    df_grouped = saveResults.groupby("ScanType")
    periods = configManager.periodsRange
    sum_list = []
    sum_dict = {}
    maxGrowth = -100
    for scanType, df_group in df_grouped:
        groupItems = len(df_group)
        sum_dict = {}
        sum_dict["ScanType"] = f"[SUM]{scanType.replace('(','[').replace(')',']')}  ({groupItems})"
        sum_dict["Date"] = PKDateUtilities.currentDateTime().strftime("%Y-%m-%d")
        for prd in periods:
            if not f"{prd}Pd-%" in df_group.columns:
                continue
            prd_df = df_group[[f"{prd}Pd-%", f"{prd}Pd-10k"]]
            prd_df.loc[:, f"{prd}Pd-10k"] = prd_df.loc[:, f"{prd}Pd-10k"].apply(
                lambda x: ImageUtility.PKImageTools.removeAllColorStyles(x)
            )
            prd_df = prd_df.replace("-", np.nan, regex=True)
            prd_df = prd_df.replace("", np.nan, regex=True)
            prd_df.dropna(axis=0, how="all", inplace=True)
            prd_df[f"{prd}Pd-10k"] = prd_df[f"{prd}Pd-10k"].astype(float).fillna(0.0)
            gain = round(
                (prd_df[f"{prd}Pd-10k"].sum() - 10000 * len(prd_df))
                * 100
                / (10000 * len(prd_df)),
                2,
            )
            sum_dict[f"{prd}Pd-%"] = gain
            sum_dict[f"{prd}Pd-10k"] = round(
                prd_df[f"{prd}Pd-10k"].sum() / len(prd_df), 2
            )
        sum_list.append(sum_dict)
    df = pd.DataFrame(sum_list)
    df = formatGridOutput(df, replacenan=False)
    saveResults = pd.concat([saveResults, df], axis=0)
    saveResults = saveResults.replace(np.nan, "-", regex=True)
    return saveResults


def performXRay(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        savedResults, userArgs, calcForDate, progressLabel = task.long_running_fn_args
    else:
        savedResults, userArgs, calcForDate, progressLabel = args[0],args[1],args[2],args[3]
    df = None
    if savedResults is not None and len(savedResults) > 0:
        backtestPeriods = getbacktestPeriod(userArgs)
        saveResults = cleanupData(savedResults)

        days = 0
        periods = configManager.periodsRange
        # period = periods[days]
        # requiredPeriods = []
        # If the calcForDate is more than the backtestPeriods, we should be able to 
        # still calculate the backtested returns for all 30 periods or periods more than
        # the requested backtestPeriods
        # backtestPeriods = getUpdatedBacktestPeriod(calcForDate, backtestPeriods, saveResults)
        # while periods[days] <= backtestPeriods:
        #     period = periods[days]
        #     requiredPeriods.append(period)
        #     days += 1
        #     if days >= len(periods):
        #         break
        df = getBacktestDataFromCleanedData(userArgs, saveResults, df, periods,progressLabel)
        if df is None:
            return None
        df = cleanFormattingForStatsData(calcForDate, saveResults, df)
        # OutputControls().printOutput(f"All portfolios:\n{PortfolioCollection().portfoliosAsDataframe}")
        # OutputControls().printOutput(f"All portfoliosSummary:\n{PortfolioCollection().ledgerSummaryAsDataframe}")
    if task is not None:
        if task.taskId > 0:
            task.progressStatusDict[task.taskId] = {'progress': 0, 'total': 1}
            task.resultsDict[task.taskId] = df
        else:
            task.result = df
    return df

def getUpdatedBacktestPeriod(calcForDate, backtestPeriods, saveResults):
    targetDate = (
            calcForDate if calcForDate is not None else saveResults["Date"].iloc[0]
        )
    today = PKDateUtilities.currentDateTime()
    gap = PKDateUtilities.trading_days_between(
            PKDateUtilities.dateFromYmdString(targetDate)
            .replace(tzinfo=today.tzinfo)
            .date(),
            today.date(),
        )
    backtestPeriods = gap if gap > backtestPeriods else backtestPeriods
    return backtestPeriods if backtestPeriods <= configManager.maxBacktestWindow else configManager.maxBacktestWindow

def cleanFormattingForStatsData(calcForDate, saveResults, df):
    if df is None or not isinstance(df, pd.DataFrame) or df.empty \
        or saveResults is None or not isinstance(saveResults, pd.DataFrame) or saveResults.empty:
        return df
    df = df[
            [
                col
                for col in df.columns
                if ("ScanType" in col or "Pd-%" in col or "Pd-10k" in col)
            ]
        ]
    df = df.replace(999999999, np.nan, regex=True)
    df.dropna(axis=0, how="all", inplace=True)
    df = formatGridOutput(df)
    df.insert(
            1,
            "Date",
            calcForDate if calcForDate is not None else saveResults["Date"].iloc[0],
        )
    
    return df

def getBacktestDataFromCleanedData(args, saveResults, df, periods,progressLabel:str=None):
    '''
    Important
    ---------
    You should have called `cleanupData` before calling this.
    '''
    for period in periods:
        saveResults[f"LTP{period}"] = (
                    saveResults[f"LTP{period}"].astype(float).fillna(0.0)
                )
        saveResults[f"Growth{period}"] = (
                    saveResults[f"Growth{period}"].astype(float).fillna(0.0)
                )

    scanResults = statScanCalculations(args, saveResults, periods,progressLabel)

    if df is None:
        df = pd.DataFrame(scanResults)
    else:
        df1 = pd.DataFrame(scanResults)
        df_target = df1[
                    [col for col in df1.columns if ("Pd-%" in col or "Pd-10k" in col)]
                ]
        df = pd.concat([df, df_target], axis=1)
    return df

def statScanCalculationForNoFilter(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(saveResults, period, "NoFilter", userArgs,task)
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForPatterns(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    df_grouped = saveResults.groupby("Pattern")
    for pattern, df_group in df_grouped:
        if pattern is None or len(pattern) == 0:
            pattern = "No Pattern"
        scanResults.append(
                    getCalculatedValues(df_group, period, f"[P]{pattern}", userArgs,task)
                )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def ensureColumnsExist(saveResults):
    columns = ['Stock', 'Date', "volume", 'Trend', 'MA-Signal', 'LTP', '52Wk-H',
               '52Wk-L', '1-Pd', '2-Pd', '3-Pd', '4-Pd', '5-Pd', '10-Pd', '15-Pd',
               '22-Pd', '30-Pd', 'Consol.', 'Breakout', 'RSI', 'Pattern', 'CCI',
               'LTP1', 'Growth1', 'LTP2', 'Growth2', 'LTP3', 'Growth3', 'LTP4',
               'Growth4', 'LTP5', 'Growth5', 'LTP10', 'Growth10', 'LTP15', 'Growth15',
               'LTP22', 'Growth22', 'LTP30', 'Growth30']
    if saveResults is None:
        saveResults = pd.DataFrame(columns=columns)
    else:
        for col in columns:
            if col not in saveResults.columns:
                saveResults[col] = ""
    return saveResults

def cleanupData(savedResults):
    saveResults = savedResults.copy()
    saveResults = ensureColumnsExist(saveResults)
    for col in saveResults.columns:
        saveResults.loc[:, col] = saveResults.loc[:, col].apply(
                lambda x: ImageUtility.PKImageTools.removeAllColorStyles(x)
            )

    saveResults["LTP"] = saveResults["LTP"].astype(float).fillna(0.0)
    saveResults["RSI"] = saveResults["RSI"].astype(float).fillna(0.0)
    saveResults.loc[:, "volume"] = saveResults.loc[:, "volume"].apply(
            lambda x: x.replace("x", "")
        )
    if f"Trend({configManager.daysToLookback}Prds)" not in saveResults.columns:
        saveResults.rename(
                columns={
                    # "Consol.": f"Consol.({configManager.daysToLookback}Prds)",
                    "Trend": f"Trend({configManager.daysToLookback}Prds)",
                    "Breakout": f"Breakout({configManager.daysToLookback}Prds)",
                },
                inplace=True,
            )
    saveResults.loc[:, f"Consol."] = saveResults.loc[
            :, f"Consol."
        ].apply(lambda x: x.replace("Range:", "").replace("%", ""))
    try:
        saveResults = saveResults.loc[:,(saveResults!='').any(axis=0)]
    except ValueError:
        # The truth value of a Series is ambiguous.
        pass
    if f"Breakout({configManager.daysToLookback}Prds)" in savedResults.columns:
        saveResults[["Breakout", "Resistance"]] = saveResults[
                f"Breakout({configManager.daysToLookback}Prds)"
            ].astype(str).str.split(" R: ", n=1, expand=True)
        saveResults.loc[:, "Breakout"] = saveResults.loc[:, "Breakout"].apply(
                lambda x: x.replace("BO: ", "").replace(" ", "")
            )
        saveResults.loc[:, "Resistance"] = saveResults.loc[
                :, "Resistance"
            ].apply(lambda x: x.replace("(Potential)", "") if x is not None else x)
        saveResults["Breakout"] = saveResults["Breakout"].astype(float).fillna(0.0)
        saveResults["Resistance"] = saveResults["Resistance"].astype(float).fillna(0.0)
    saveResults["volume"] = saveResults["volume"].astype(float).fillna(0.0)
    saveResults[f"Consol."] = (
            saveResults[f"Consol."].astype(float).fillna(0.0)
        )
    saveResults["52Wk-H"] = saveResults["52Wk-H"].astype(float).fillna(0.0)
    saveResults["52Wk-L"] = saveResults["52Wk-L"].astype(float).fillna(0.0)
    saveResults["CCI"] = saveResults["CCI"].astype(float).fillna(0.0)
    return saveResults

def getbacktestPeriod(args):
    backtestPeriods = configManager.maxBacktestWindow  # Max backtest days
    if args is None or ((not isinstance(args,int)) and (not isinstance(args,Namespace))):
        return backtestPeriods
    if args is not None and args.backtestdaysago is not None:
        try:
            backtestPeriods = int(args.backtestdaysago)
        except Exception as e:# pragma: no cover
            default_logger().debug(e, exc_info=True)
            pass
    return backtestPeriods

def statScanCalculations(userArgs, saveResults, periods,progressLabel:str=None):
    scanResults = []
    tasksList = []
    if saveResults is not None and len(saveResults) >= 1:
        task1 = PKTask(f"[{len(saveResults)}] RSI Stats",long_running_fn=statScanCalculationForRSI)
        task2 = PKTask(f"[{len(saveResults)}] Trend Stats",long_running_fn=statScanCalculationForTrend)
        task3 = PKTask(f"[{len(saveResults)}] MA Stats",long_running_fn=statScanCalculationForMA)
        task4 = PKTask(f"[{len(saveResults)}] Volume Stats",long_running_fn=statScanCalculationForVol)
        task5 = PKTask(f"[{len(saveResults)}] Consolidation Stats",long_running_fn=statScanCalculationForConsol)
        task6 = PKTask(f"[{len(saveResults)}] Breakout Stats",long_running_fn=statScanCalculationForBO)
        task7 = PKTask(f"[{len(saveResults)}] 52Week Stats",long_running_fn=statScanCalculationFor52Wk)
        task8 = PKTask(f"[{len(saveResults)}] CCI Stats",long_running_fn=statScanCalculationForCCI)
        task9 = PKTask(f"[{len(saveResults)}] CCI Stats",long_running_fn=statScanCalculationForPatterns)
        task10 = PKTask(f"[{len(saveResults)}] NoFilter Stats",long_running_fn=statScanCalculationForNoFilter)
        tasksList=[task1,task2,task3,task4,task5,task6,task7,task8,task9,task10]
        for task in tasksList:
            task.long_running_fn_args = (userArgs, saveResults, periods, scanResults)
        if configManager.enablePortfolioCalculations:
            PKScheduler.scheduleTasks(tasksList,label=progressLabel,showProgressBars=True,timeout=600)
        else:
            for task in tasksList:
                task.long_running_fn(*(task,))
        for task in tasksList:
            if task.result is not None and len(task.result) > 0:
                scanResults.extend(task.result)
    return scanResults

def statScanCalculationForCCI(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterCCIBelowMinus100(saveResults), period, "[CCI]<=-100", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterCCIBelow0(saveResults), period, "[CCI]-100<C<0", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterCCI0To100(saveResults), period, "[CCI]0<=C<=100", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterCCI100To200(saveResults), period, "[CCI]100<C<=200", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterCCIAbove200(saveResults), period, "[CCI]>200", userArgs,task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationFor52Wk(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterLTPMoreOREqual52WkH(saveResults), period, "[52Wk]LTP>=H", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPWithin90Percent52WkH(saveResults),
                    period,
                    "[52Wk]LTP>=.9*H",
                    userArgs,
                    task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPLess90Percent52WkH(saveResults),
                    period,
                    "[52Wk]LTP<.9*H",
                    userArgs,
                    task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPMore52WkL(saveResults), period, "[52Wk]LTP>L", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPWithin90Percent52WkL(saveResults),
                    period,
                    "[52Wk]LTP>=1.1*L",
                    userArgs,
                    task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPLess52WkL(saveResults), period, "[52Wk]LTP<=L", userArgs,task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForBO(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterLTPLessThanBreakout(saveResults), period, "[BO]LTP<BO", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPMoreOREqualBreakout(saveResults),
                    period,
                    "[BO]LTP>=BO",
                    userArgs,
                    task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPLessThanResistance(saveResults), period, "[BO]LTP<R", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterLTPMoreOREqualResistance(saveResults),
                    period,
                    "[BO]LTP>=R",
                    userArgs,
                    task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForConsol(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterConsolidating10Percent(saveResults), period, "Cons.<=10", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterConsolidatingMore10Percent(saveResults),
                    period,
                    "Cons.>10",
                    userArgs,
                    task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForVol(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterVolumeLessThan25(saveResults), period, "Vol<2.5", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterVolumeMoreThan25(saveResults), period, "Vol>=2.5", userArgs,task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForMA(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterMASignalBullish(saveResults), period, "[MA]Bull", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterMASignalBearish(saveResults), period, "[MA]Bear", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterMASignalNeutral(saveResults), period, "[MA]Neutral", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterMASignalBullCross(saveResults), period, "[MA]BullCross", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterMASignalBearCross(saveResults), period, "[MA]BearCross", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterMASignalSupport(saveResults), period, "[MA]Support", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterMASignalResist(saveResults), period, "[MA]Resist", userArgs,task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForTrend(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterTrendStrongUp(saveResults), period, "[T]StrongUp", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterTrendWeakUp(saveResults), period, "[T]WeakUp", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterTrendUp(saveResults), period, "[T]TrendUp", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterTrendStrongDown(saveResults), period, "[T]StrongDown", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterTrendWeakDown(saveResults), period, "[T]WeakDown", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterTrendSideways(saveResults), period, "[T]Sideways", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterTrendDown(saveResults), period, "[T]TrendDown", userArgs,task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults

def statScanCalculationForRSI(*args, **kwargs):
    task = None
    if isinstance(args[0], PKTask):
        task = args[0]
        userArgs, saveResults, period, scanResults = task.long_running_fn_args
    else:
        userArgs, saveResults, period, scanResults = args[0],args[1],args[2],args[3]
    scanResults.append(
                getCalculatedValues(
                    filterRSIAbove50(saveResults), period, "[RSI]>=50", userArgs, task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterRSI50To67(saveResults), period, "[RSI]50<=RSI<=67", userArgs,task
                )
            )
    scanResults.append(
                getCalculatedValues(
                    filterRSI68OrAbove(saveResults), period, "[RSI]>=68", userArgs,task
                )
            )
    if task is not None:
        if task.taskId > 0:
            task.resultsDict[task.taskId] = scanResults
            task.progressStatusDict[task.taskId] = {'progress': 1, 'total': 1}
        else:
            task.result = scanResults
    return scanResults


def formatGridOutput(df,replacenan=True):
    if replacenan:
        df = df.replace(np.nan, "-", regex=True)
    for col in df.columns:
        try:
            df[col] = df[col].astype(float).fillna(0)
        except ValueError as e: # pragma: no cover
            if len(str(e).split("-")) < 2:
                # We want to allow value "-" in column values
                # We don't want to allow '2024-12-03' or 'RSI >=50' or "Trends-50-to-60"
                pass
            else:
                continue
        except Exception as e:# pragma: no cover
            default_logger().debug(f"For column:{col}:\n{e}", exc_info=True)
            continue
        maxGrowth = df[col].max()
        if "Pd-%" in col:
            df.loc[:, col] = df.loc[:, col].astype(str).apply(
                lambda x: x
                if (str(x) == "-")
                else (
                    str(x).replace(
                        str(x),
                        (
                            (
                                (colorText.WHITE)
                                if x == maxGrowth
                                else colorText.GREEN
                            )
                            if float(x) > 0
                            else (colorText.FAIL if float(x) < 0 else colorText.WARN)
                        )
                        + str(float(x))
                        + " %"
                        + colorText.END,
                    )
                )
            )
        if "Pd-10k" in col:
            df.loc[:, col] = df.loc[:, col].astype(str).apply(
                lambda x: x
                if (str(x) == "-")
                else (
                    str(x).replace(
                        str(x),
                        (
                            (
                                (colorText.WHITE)
                                if x == maxGrowth
                                else colorText.GREEN
                            )
                            if (float(x) > 10000)
                            else (
                                colorText.FAIL if float(x) < 10000 else colorText.WARN
                            )
                        )
                        + str(x)
                        + colorText.END,
                    )
                )
            )
    df = df.replace(np.nan, "-", regex=True)
    df = df.replace(colorText.FAIL + "-100.0 %" + colorText.END, "-", regex=False)
    df = df.replace(colorText.WARN + "0.0 %" + colorText.END, "-", regex=False)
    df = df.replace(colorText.FAIL + "0.0" + colorText.END, "-", regex=False)
    return df


def getCalculatedValues(df, periods, key, userArgs=None, task=None):
    collated_df = None
    for period in periods:
        ltpSum1ShareEach = round(df["LTP"].sum(), 2)
        tdySum1ShareEach = round(df[f"LTP{period}"].sum(), 2)
        growthSum1ShareEach = round(df[f"Growth{period}"].sum(), 2)
        percentGrowth = 0
        if ltpSum1ShareEach > 0:
            percentGrowth = round(100 * growthSum1ShareEach / ltpSum1ShareEach, 2)
        growth10k = round(10000 * (1 + 0.01 * percentGrowth), 2)
        if configManager.enablePortfolioCalculations and userArgs.options.startswith("B"): # backtests
            portfolio = Portfolio(name=key)
            portfolio.updatePortfolioFromXRayDataFrame(df,configManager.periodsRange,task)
            PortfolioCollection().addPortfolio(portfolio)
        result_df = {
            "ScanType": key, #if tdySum1ShareEach != 0 else 999999999,
            f"{period}Pd-PFV": tdySum1ShareEach,
            f"{period}Pd-%": percentGrowth if tdySum1ShareEach != 0 else 999999999,
            f"{period}Pd-10k": growth10k if tdySum1ShareEach != 0 else 999999999,
        }
        if collated_df is None:
            collated_df = result_df
        collated_df = collated_df | result_df
    # percentGrowth = colorText.GREEN if percentGrowth >=0 else colorText.FAIL + percentGrowth + colorText.END
    # growth10k = colorText.GREEN if percentGrowth >=0 else colorText.FAIL + growth10k + colorText.END
    # df_col = {'ScanType':key,
    #     f'{period}Pd-PFV':tdySum1ShareEach,
    #     f'{period}Pd-PFG':percentGrowth if tdySum1ShareEach != 0 else '-',
    #     f'{period}Pd-Go10k':growth10k if tdySum1ShareEach != 0 else '-',
    #     }

    return collated_df  # , df_col


def filterRSIAbove50(df):
    if df is None:
        return None
    return df[df["RSI"] > 50].fillna(0.0)


def filterRSI50To67(df):
    if df is None:
        return None
    return df[(df["RSI"] >= 50) & (df["RSI"] <= 67)].fillna(0.0)


def filterRSI68OrAbove(df):
    if df is None:
        return None
    return df[df["RSI"] >= 68].fillna(0.0)


def filterTrendStrongUp(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"] == "Strong Up"].fillna(0.0)


def filterTrendWeakUp(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"] == "Weak Up"].fillna(0.0)


def filterTrendWeakDown(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"] == "Weak Down"].fillna(0.0)


def filterTrendStrongDown(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"] == "Strong Down"].fillna(0.0)


def filterTrendUp(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"].astype(str).str.endswith("Up")].fillna(0.0)


def filterTrendSideways(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"] == "Sideways"].fillna(0.0)


def filterTrendDown(df):
    if df is None:
        return None
    return df[df[f"Trend({configManager.daysToLookback}Prds)"].astype(str).str.endswith("Down")].fillna(0.0)


def filterMASignalBullish(df):
    if df is None:
        return None
    return df[df["MA-Signal"] == "Bullish"].fillna(0.0)


def filterMASignalBearish(df):
    if df is None:
        return None
    return df[df["MA-Signal"] == "Bearish"].fillna(0.0)


def filterMASignalNeutral(df):
    if df is None:
        return None
    return df[df["MA-Signal"] == "Neutral"].fillna(0.0)


def filterMASignalBullCross(df):
    if df is None:
        return None
    return df[df["MA-Signal"].astype(str).str.startswith("BullCross")].fillna(0.0)


def filterMASignalBearCross(df):
    if df is None:
        return None
    return df[df["MA-Signal"].astype(str).str.startswith("BearCross")].fillna(0.0)


def filterMASignalSupport(df):
    if df is None:
        return None
    return df[df["MA-Signal"].astype(str).str.endswith("Support")].fillna(0.0)


def filterMASignalResist(df):
    if df is None:
        return None
    return df[df["MA-Signal"].astype(str).str.endswith("Resist")].fillna(0.0)


def filterVolumeLessThan25(df):
    if df is None:
        return None
    return df[df["volume"] < 2.5].fillna(0.0)


def filterVolumeMoreThan25(df):
    if df is None:
        return None
    return df[df["volume"] >= 2.5].fillna(0.0)


def filterConsolidating10Percent(df):
    if df is None:
        return None
    return df[df[f"Consol."] <= 10].fillna(0.0)


def filterConsolidatingMore10Percent(df):
    if df is None:
        return None
    return df[df[f"Consol."] > 10].fillna(0.0)


def filterLTPLessThanBreakout(df):
    if df is None:
        return None
    return df[df["LTP"] < df["Breakout"]].fillna(0.0)


def filterLTPMoreOREqualBreakout(df):
    if df is None:
        return None
    return df[((df["Breakout"] > 0) & (df["LTP"] >= df["Breakout"]))].fillna(0.0)


def filterLTPLessThanResistance(df):
    if df is None:
        return None
    return df[df["LTP"] < df["Resistance"]].fillna(0.0)


def filterLTPMoreOREqualResistance(df):
    if df is None:
        return None
    return df[((df["Resistance"] > 0) & (df["LTP"] >= df["Resistance"]))].fillna(0.0)


def filterLTPMoreOREqual52WkH(df):
    if df is None:
        return None
    return df[df["LTP"] >= df["52Wk-H"]].fillna(0.0)


def filterLTPWithin90Percent52WkH(df):
    if df is None:
        return None
    return df[(df["LTP"] >= 0.9 * df["52Wk-H"]) & (df["LTP"] < df["52Wk-H"])].fillna(
        0.0
    )


def filterLTPLess90Percent52WkH(df):
    if df is None:
        return None
    return df[df["LTP"] < 0.9 * df["52Wk-H"]].fillna(0.0)


def filterLTPMore52WkL(df):
    if df is None:
        return None
    return df[((df["LTP"] > df["52Wk-L"]) & (df["LTP"] < 1.1 * df["52Wk-L"]))].fillna(
        0.0
    )


def filterLTPWithin90Percent52WkL(df):
    if df is None:
        return None
    return df[(df["LTP"] >= (1.1 * df["52Wk-L"])) & (df["LTP"] > df["52Wk-L"])].fillna(
        0.0
    )


def filterLTPLess52WkL(df):
    if df is None:
        return None
    return df[df["LTP"] <= df["52Wk-L"]].fillna(0.0)


def filterCCIBelowMinus100(df):
    if df is None:
        return None
    return df[df["CCI"] <= -100].fillna(0.0)


def filterCCIBelow0(df):
    if df is None:
        return None
    return df[(df["CCI"] > -100) & (df["CCI"] < 0)].fillna(0.0)


def filterCCI0To100(df):
    if df is None:
        return None
    return df[(df["CCI"] >= 0) & (df["CCI"] <= 100)].fillna(0.0)


def filterCCI100To200(df):
    if df is None:
        return None
    return df[(df["CCI"] > 100) & (df["CCI"] <= 200)].fillna(0.0)


def filterCCIAbove200(df):
    if df is None:
        return None
    return df[(df["CCI"] > 200)].fillna(0.0)

def returnNoFilter(df):
    return df

def filterPattern(df, pattern="[P]No Pattern"):
    if df is None:
        return None
    match_df = None
    df_grouped = df.groupby("Pattern")
    for foundPattern, df_group in df_grouped:
        if foundPattern is None or len(foundPattern) == 0:
            foundPattern = "No Pattern"
        if f"[P]{foundPattern}" == pattern:
            match_df = df_group
            break
    return match_df

def strategyDictionary():
    """
    Available Strategies:

    RSI
    ---
    `RSI>=50`, `50<=RSI<=67` and `RSI>=68`

    CCI
    ---
    `[CCI]<=-100`, `[CCI]-100<C<0`, `[CCI]0<=C<=100`, `[CCI]100<C<=200` and `[CCI]>200`

    52Wk-H/L
    --------
    `[52Wk]LTP>=H`, `[52Wk]LTP>=.9*H`, `[52Wk]LTP<.9*H`, `[52Wk]LTP>L`, `[52Wk]LTP>=1.1*L` and `[52Wk]LTP<=L`

    Breakout
    --------
    `[BO]LTP<BO`, `[BO]LTP>=BO`, `[BO]LTP<R` and `[BO]LTP>=R`

    Consolidation
    -------------
    `Cons.<=10` and Cons.>10

    Volume
    ------
    `Vol<2.5` and `Vol>=2.5`

    Moving Averages
    ---------------
    `[MA]Bull`, `[MA]Bear`, `[MA]Neutral`, `[MA]BullCross`, `[MA]BearCross`, `[MA]Support` and `[MA]Resist`

    Trends
    ------
    `[T]StrongUp`, `[T]WeakUp`, `[T]TrendUp`, `[T]StrongDown`, `[T]WeakDown`, `[T]Sideways` and `[T]TrendDown`

    NoFilter
    --------
    `NoFilter`

    Patterns
    --------
    `[P]{patern_name}`. For example, `[P]No Pattern`
    """
    strategies = {}
    # RSI
    strategies["[RSI]>=50"] = filterRSIAbove50
    strategies["[RSI]50<=RSI<=67"] = filterRSI50To67
    strategies["[RSI]>=68"] = filterRSI68OrAbove
    # CCI
    strategies["[CCI]<=-100"] = filterCCIBelowMinus100
    strategies["[CCI]-100<C<0"] = filterCCIBelow0
    strategies["[CCI]0<=C<=100"] = filterCCI0To100
    strategies["[CCI]100<C<=200"] = filterCCI100To200
    strategies["[CCI]>200"] = filterCCIAbove200
    # 52Wk-H/L
    strategies["[52Wk]LTP>=H"] = filterLTPMoreOREqual52WkH
    strategies["[52Wk]LTP>=.9*H"] = filterLTPWithin90Percent52WkH
    strategies["[52Wk]LTP<.9*H"] = filterLTPLess90Percent52WkH
    strategies["[52Wk]LTP>L"] = filterLTPMore52WkL
    strategies["[52Wk]LTP>=1.1*L"] = filterLTPWithin90Percent52WkL
    strategies["[52Wk]LTP<=L"] = filterLTPLess52WkL
    # Breakout
    strategies["[BO]LTP<BO"] = filterLTPLessThanBreakout
    strategies["[BO]LTP>=BO"] = filterLTPMoreOREqualBreakout
    strategies["[BO]LTP<R"] = filterLTPLessThanResistance
    strategies["[BO]LTP>=R"] = filterLTPMoreOREqualResistance
    # Consolidation
    strategies["Cons.<=10"] = filterConsolidating10Percent
    strategies["Cons.>10"] = filterConsolidatingMore10Percent
    # Volume
    strategies["Vol<2.5"] = filterVolumeLessThan25
    strategies["Vol>=2.5"] = filterVolumeMoreThan25
    # MA
    strategies["[MA]Bull"] = filterMASignalBullish
    strategies["[MA]Bear"] = filterMASignalBearish
    strategies["[MA]Neutral"] = filterMASignalNeutral
    strategies["[MA]BullCross"] = filterMASignalBullCross
    strategies["[MA]BearCross"] = filterMASignalBearCross
    strategies["[MA]Support"] = filterMASignalSupport
    strategies["[MA]Resist"] = filterMASignalResist
    # Trends
    strategies["[T]StrongUp"] = filterTrendStrongUp
    strategies["[T]WeakUp"] = filterTrendWeakUp
    strategies["[T]TrendUp"] = filterTrendUp
    strategies["[T]StrongDown"] = filterTrendStrongDown
    strategies["[T]WeakDown"] = filterTrendWeakDown
    strategies["[T]Sideways"] = filterTrendSideways
    strategies["[T]TrendDown"] = filterTrendDown
    # NoFilter
    strategies["NoFilter"] = returnNoFilter
    # Pattern
    strategies["[P]"] = filterPattern
    return strategies

def strategyForKey(key:str):
    strategies = strategyDictionary()
    if key in strategies.keys():
        return strategies[key]
    elif key.startswith("[P]"):
        return strategies["[P]"]

def strategyNames():
    return strategyDictionary().keys()


#!/usr/bin/env python
"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import copy
import datetime
import shutil
import sys
import os
import numpy as np
import pandas as pd
import pkscreener.classes.Utility as Utility
from pkscreener.classes.ConfigManager import parser, tools
from pkscreener.classes.ScreeningStatistics import ScreeningStatistics
from pkscreener.classes import AssetsManager

from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes import Archiver
from PKDevTools.classes.Singleton import SingletonType, SingletonMixin
from PKDevTools.classes.PKPickler import PKPicklerDB
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.MarketHours import MarketHours
from PKDevTools.classes.log import default_logger
from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.SuppressOutput import SuppressOutput
from PKDevTools.classes.MarketHours import MarketHours

from halo import Halo

configManager = tools()

STD_ENCODING=sys.stdout.encoding if sys.stdout is not None else 'utf-8'

class PKDailyStockDataDB(SingletonMixin, metaclass=SingletonType):
    def __init__(self,fileName=None):
        super(PKDailyStockDataDB, self).__init__()
        self.pickler = PKPicklerDB(fileName=fileName)

    def searchCache(self, ticker:str=None, name:str=None):
        return self.pickler.searchCache(ticker=ticker, name=name)
    
    def saveCache(self, ticker:str=None, name:str=None, stockDict:dict=None):
        self.pickler.saveCache(ticker=ticker, name=name, stockDict=stockDict)

class PKIntradayStockDataDB(SingletonMixin, metaclass=SingletonType):
    def __init__(self,fileName=None):
        super(PKIntradayStockDataDB, self).__init__()
        self.pickler = PKPicklerDB(fileName=fileName)

    def searchCache(self, ticker:str=None, name:str=None):
        return self.pickler.searchCache(ticker=ticker, name=name)
    
    def saveCache(self, ticker:str=None, name:str=None, stockDict:dict=None):
        self.pickler.saveCache(ticker=ticker, name=name, stockDict=stockDict)


class PKMarketOpenCloseAnalyser:
    configManager = tools()
    configManager.getConfig(parser)
    updatedCandleData = None
    allDailyCandles = None
    allIntradayCandles = None
    
    def getStockDataForSimulation(sliceWindowDatetime=None,listStockCodes=[]):
        int_exists, int_cache_file, stockDictInt = PKMarketOpenCloseAnalyser.ensureIntradayStockDataExists(listStockCodes=listStockCodes)
        daily_exists, daily_cache_file, stockDict = PKMarketOpenCloseAnalyser.ensureDailyStockDataExists(listStockCodes=listStockCodes)
        updatedCandleData = PKMarketOpenCloseAnalyser.updatedCandleData
        allDailyCandles = PKMarketOpenCloseAnalyser.allDailyCandles
        if updatedCandleData is not None and len(updatedCandleData) < 1:
            updatedCandleData = None
        if allDailyCandles is not None and len(allDailyCandles) < 1:
            allDailyCandles = None
        if  ((int_exists or len(stockDictInt) > 0) and (daily_exists or len(stockDict) > 0)) and (updatedCandleData is None or allDailyCandles is None):
            allDailyCandles = PKMarketOpenCloseAnalyser.getLatestDailyCandleData(daily_cache_file,stockDict)
            morningIntradayCandle = PKMarketOpenCloseAnalyser.getIntradayCandleFromMorning(int_cache_file,sliceWindowDatetime=sliceWindowDatetime,stockDictInt=stockDictInt)
            updatedCandleData = PKMarketOpenCloseAnalyser.combineDailyStockDataWithMorningSimulation(allDailyCandles,morningIntradayCandle)
            PKMarketOpenCloseAnalyser.updatedCandleData = updatedCandleData
            PKMarketOpenCloseAnalyser.allDailyCandles = allDailyCandles
            AssetsManager.PKAssetsManager.saveStockData(updatedCandleData,PKMarketOpenCloseAnalyser.configManager,1,False,False, True)
        return updatedCandleData, allDailyCandles

    @Halo(text='  [+] Running final analysis...', spinner='dots')
    def runOpenCloseAnalysis(updatedCandleData,allDailyCandles,screen_df,save_df,runOptionName=None,filteredListOfStocks=[]):
        # stockListFromMorningTrade,morningIntraday_df = PKMarketOpenCloseAnalyser.simulateMorningTrade(updatedCandleData)
        # latest_daily_df = PKMarketOpenCloseAnalyser.runScanForStocksFromMorningTrade(stockListFromMorningTrade,allDailyCandles)
        try:
            shouldSuppress = not OutputControls().enableMultipleLineOutput
            with SuppressOutput(suppress_stderr=shouldSuppress, suppress_stdout=shouldSuppress):
                save_df, screen_df = PKMarketOpenCloseAnalyser.diffMorningCandleDataWithLatestDailyCandleData(screen_df,save_df, updatedCandleData, allDailyCandles,runOptionName=runOptionName,filteredListOfStocks=filteredListOfStocks)
        except: # pragma: no cover
            pass
        AssetsManager.PKAssetsManager.saveStockData(allDailyCandles,PKMarketOpenCloseAnalyser.configManager,1,False,False, True)
        return save_df, screen_df

    @Halo(text='  [+] Getting intraday data...', spinner='dots')
    def ensureIntradayStockDataExists(listStockCodes=[]):
        # Ensure that the intraday_stock_data_<date>.pkl file exists
        exists, cache_file = AssetsManager.PKAssetsManager.afterMarketStockDataExists(intraday=True)
        copyFilePath = os.path.join(Archiver.get_user_data_dir(), f"copy_{cache_file}")
        srcFilePath = os.path.join(Archiver.get_user_data_dir(), cache_file)
        srcFileSize = os.stat(srcFilePath).st_size if os.path.exists(srcFilePath) else 0
        stockDict = None
        if exists and srcFileSize < 1024*1024*40:
             # File less than 30MB ? Must have been corrupted
            try:
                os.remove(srcFilePath)
                exists = False
            except: # pragma: no cover
                pass
        isTrading = PKDateUtilities.isTradingTime()
        if not exists or isTrading:
            savedPeriod = PKMarketOpenCloseAnalyser.configManager.period
            savedDuration = PKMarketOpenCloseAnalyser.configManager.duration
            PKMarketOpenCloseAnalyser.configManager.period = "1d"
            PKMarketOpenCloseAnalyser.configManager.duration = "1m"
            PKMarketOpenCloseAnalyser.configManager.setConfig(parser, default=True, showFileCreatedText=False)
            OutputControls().printOutput(f"  [+] {colorText.FAIL}{cache_file}{colorText.END} not found under {Archiver.get_user_data_dir()} !")
            OutputControls().printOutput(f"  [+] {colorText.GREEN}Trying to download {cache_file}{colorText.END}. Please wait ...")
            if os.path.exists(copyFilePath) and not isTrading:
                copyFileSize = os.stat(copyFilePath).st_size if os.path.exists(copyFilePath) else 0
                if copyFileSize >= 1024*1024*40:
                    shutil.copy(copyFilePath,srcFilePath) # copy is the saved source of truth
                    PKMarketOpenCloseAnalyser.configManager.period = savedPeriod
                    PKMarketOpenCloseAnalyser.configManager.duration = savedDuration
                    PKMarketOpenCloseAnalyser.configManager.setConfig(parser, default=True, showFileCreatedText=False)
                    return True, cache_file, stockDict
            stockDict = AssetsManager.PKAssetsManager.loadStockData(stockDict={},configManager=PKMarketOpenCloseAnalyser.configManager,downloadOnly=False,defaultAnswer='Y',retrial=False,forceLoad=False,stockCodes=listStockCodes,isIntraday=True)
            exists, cache_file = AssetsManager.PKAssetsManager.afterMarketStockDataExists(intraday=True)
            PKMarketOpenCloseAnalyser.configManager.period = savedPeriod
            PKMarketOpenCloseAnalyser.configManager.duration = savedDuration
            PKMarketOpenCloseAnalyser.configManager.setConfig(parser, default=True, showFileCreatedText=False)
            if not exists and len(stockDict) <= 0:
                OutputControls().printOutput(f"  [+] {colorText.FAIL}{cache_file}{colorText.END} not found under {Archiver.get_user_data_dir()}/ !")
                OutputControls().printOutput(f"  [+] Please run {colorText.FAIL}pkscreener{colorText.END}{colorText.GREEN} -a Y -e -d -i 1m{colorText.END} and then run this menu option again.")
                OutputControls().takeUserInput("Press any key to continue...")
        try:
            if os.path.exists(copyFilePath) and exists:
                shutil.copy(copyFilePath,srcFilePath) # copy is the saved source of truth
            if not os.path.exists(copyFilePath) and exists: # Let's make a copy of the original one
                shutil.copy(srcFilePath,copyFilePath)
        except: # pragma: no cover
            pass
        return exists, cache_file, stockDict

    @Halo(text='  [+] Getting daily data...', spinner='dots')
    def ensureDailyStockDataExists(listStockCodes=[]):
        # Ensure that the stock_data_<date>.pkl file exists
        exists, cache_file = AssetsManager.PKAssetsManager.afterMarketStockDataExists(intraday=False)
        copyFilePath = os.path.join(Archiver.get_user_data_dir(), f"copy_{cache_file}")
        srcFilePath = os.path.join(Archiver.get_user_data_dir(), cache_file)
        srcFileSize = os.stat(srcFilePath).st_size if os.path.exists(srcFilePath) else 0
        stockDict = None
        if exists and srcFileSize < 1024*1024*40:
             # File less than 30MB ? Must have been corrupted
            try:
                os.remove(srcFilePath)
                exists = False
            except: # pragma: no cover
                pass
        isTrading = PKDateUtilities.isTradingTime()
        if not exists or isTrading:
            savedPeriod = PKMarketOpenCloseAnalyser.configManager.period
            savedDuration = PKMarketOpenCloseAnalyser.configManager.duration
            PKMarketOpenCloseAnalyser.configManager.period = "1y"
            PKMarketOpenCloseAnalyser.configManager.duration = "1d"
            PKMarketOpenCloseAnalyser.configManager.setConfig(parser, default=True, showFileCreatedText=False)
            OutputControls().printOutput(f"  [+] {colorText.FAIL}{cache_file}{colorText.END} not found under {Archiver.get_user_data_dir()} !")
        # We should download a fresh copy anyways because we may have altered the existing copy in
        # the previous run. -- !!!! Not required if we saved at the end of last operation !!!!
            OutputControls().printOutput(f"  [+] {colorText.GREEN}Trying to download {cache_file}{colorText.END}. Please wait ...")
            if os.path.exists(copyFilePath) and not isTrading:
                copyFileSize = os.stat(copyFilePath).st_size if os.path.exists(copyFilePath) else 0
                if copyFileSize >= 1024*1024*40:
                    shutil.copy(copyFilePath,srcFilePath) # copy is the saved source of truth
                    PKMarketOpenCloseAnalyser.configManager.period = savedPeriod
                    PKMarketOpenCloseAnalyser.configManager.duration = savedDuration
                    PKMarketOpenCloseAnalyser.configManager.setConfig(parser, default=True, showFileCreatedText=False)
                    return True, cache_file, stockDict
            stockDict = AssetsManager.PKAssetsManager.loadStockData(stockDict={},configManager=PKMarketOpenCloseAnalyser.configManager,downloadOnly=False,defaultAnswer='Y',retrial=False,forceLoad=False,stockCodes=listStockCodes,isIntraday=False,forceRedownload=True)
            exists, cache_file = AssetsManager.PKAssetsManager.afterMarketStockDataExists(intraday=False)
            PKMarketOpenCloseAnalyser.configManager.period = savedPeriod
            PKMarketOpenCloseAnalyser.configManager.duration = savedDuration
            PKMarketOpenCloseAnalyser.configManager.setConfig(parser, default=True, showFileCreatedText=False)
            if not exists and len(stockDict) <= 0:
                OutputControls().printOutput(f"  [+] {colorText.FAIL}{cache_file}{colorText.END} not found under {Archiver.get_user_data_dir()}/ !")
                OutputControls().printOutput(f"  [+] Please run {colorText.FAIL}pkscreener{colorText.END}{colorText.GREEN} -a Y -e -d{colorText.END} and then run this menu option again.")
                OutputControls().takeUserInput("Press any key to continue...")
        try:
            if os.path.exists(copyFilePath) and exists:
                shutil.copy(copyFilePath,srcFilePath) # copy is the saved source of truth
            if not os.path.exists(copyFilePath) and exists: # Let's make a copy of the original one
                shutil.copy(srcFilePath,copyFilePath)
        except: # pragma: no cover
            pass
        return exists, cache_file, stockDict
    
    def simulateMorningTrade(updatedCandleData):
        # 1. For each stock, remove the latest daily data for today from stock_data_<date>.pkl
        # 2. For each stock, only take the configManager.morninganalysiscandlenumber data rows
        #    and combine them as one candle - open for the earliest candle and close for the last candle,
        #    low and high will be the lowest and highest for in-between candles. Volume should be combined
        #    for all.
        # 3. For each stock, replace the row from #1 above with the candle data from #2 above.
        # 4. Run scan and find stocks under each (selected) scan category as if the scan was
        #    running in the morning. 
        # 5. Compare the stock prices from #4 with the removed row from #1 and show the diff.
        stockListFromMorningTrade = []
        morningIntraday_df = None
        return stockListFromMorningTrade, morningIntraday_df
    
    def getLatestDailyCandleData(daily_cache_file,stockDict=None):
        allDailyCandles = None
        if stockDict is not None and len(stockDict) > 0:
            return stockDict
        dailyDB = PKDailyStockDataDB(fileName=daily_cache_file)
        allDailyCandles = dailyDB.pickler.pickler.unpickle(fileName=dailyDB.pickler.fileName)
        # latestDailyCandle = {}
        # stocks = list(allDailyCandles.keys())
        # for stock in stocks:
        #     try:
        #         df = pd.DataFrame(data=[allDailyCandles[stock]["data"][-1]],
        #                       columns=allDailyCandles[stock]["columns"],
        #                       index=[allDailyCandles[stock]["index"][-1]])
        #         latestDailyCandle[stock] = df.to_dict("split")
        #     except: # pragma: no cover
        #         continue
        return allDailyCandles
    
    @Halo(text='  [+] Simulating morning alert...', spinner='dots')
    def getIntradayCandleFromMorning(int_cache_file=None,candle1MinuteNumberSinceMarketStarted=0,sliceWindowDatetime=None,stockDictInt=None):
        if candle1MinuteNumberSinceMarketStarted <= 0:
            candle1MinuteNumberSinceMarketStarted = PKMarketOpenCloseAnalyser.configManager.morninganalysiscandlenumber
        morningIntradayCandle = None
        if stockDictInt is not None and len(stockDictInt) > 0:
            allDailyIntradayCandles = stockDictInt
        else:
            intradayDB = PKIntradayStockDataDB(fileName=int_cache_file)
            allDailyIntradayCandles = intradayDB.pickler.pickler.unpickle(fileName=intradayDB.pickler.fileName)
        PKMarketOpenCloseAnalyser.allIntradayCandles = allDailyIntradayCandles
        morningIntradayCandle = {}
        stocks = list(allDailyIntradayCandles.keys())
        numOfCandles = PKMarketOpenCloseAnalyser.configManager.morninganalysiscandlenumber
        duration = PKMarketOpenCloseAnalyser.configManager.morninganalysiscandleduration
        numOfCandles = numOfCandles * int(duration.replace("m",""))
        for stock in stocks:
            try:
                # Let's get the saved data from the DB. Then we need to only
                # get those candles which are earlier than 9:57AM which is
                # the time when the morning alerts collect data for generating alerts
                # We'd then combine the data from 9:15 to 9:57 as a single candle of 
                # OHLCV and replace the last daily candle with this one candle to
                # simulate the scan outcome from morning.
                df = pd.DataFrame(data=allDailyIntradayCandles[stock]["data"],
                                columns=allDailyIntradayCandles[stock]["columns"],
                                index=allDailyIntradayCandles[stock]["index"])
                if sliceWindowDatetime is None:
                    df = df.head(numOfCandles)
                try:
                    alertCandleTimestamp = sliceWindowDatetime if sliceWindowDatetime is not None else f'{PKDateUtilities.tradingDate().strftime(f"%Y-%m-%d")} {MarketHours().openHour:02}:{MarketHours().openMinute+candle1MinuteNumberSinceMarketStarted}:00+05:30'
                    df = df[df.index <=  pd.to_datetime(alertCandleTimestamp).to_datetime64()]
                except: # pragma: no cover
                    alertCandleTimestamp = sliceWindowDatetime if sliceWindowDatetime is not None else f'{PKDateUtilities.tradingDate().strftime(f"%Y-%m-%d")} {MarketHours().openHour:02}:{MarketHours().openMinute+candle1MinuteNumberSinceMarketStarted}:00+05:30'
                    df = df[df.index <=  pd.to_datetime(alertCandleTimestamp, utc=True)]
                    pass
                with pd.option_context('mode.chained_assignment', None):
                    df.dropna(axis=0, how="all", inplace=True)
                if df is not None and len(df) > 0:
                    close = PKMarketOpenCloseAnalyser.getMorningClose(df)
                    adjClose = df["Adj Close"][-1] if "Adj Close" in df.columns else close
                    combinedCandle = {"open":PKMarketOpenCloseAnalyser.getMorningOpen(df), "high":max(df["high"]), 
                                    "low":min(df["low"]),"close":close,
                                    "Adj Close":adjClose,"volume":sum(df["volume"])}
                    tradingDate = df.index[-1] #PKDateUtilities.tradingDate()
                    timestamp = datetime.datetime.strptime(tradingDate.strftime("%Y-%m-%d %H:%M:%S"),"%Y-%m-%d %H:%M:%S")
                    df = pd.DataFrame([combinedCandle], columns=df.columns, index=[timestamp])
                    morningIntradayCandle[stock] = df.to_dict("split")
            except KeyboardInterrupt: # pragma: no cover
                raise KeyboardInterrupt
            except Exception as e: # pragma: no cover
                OutputControls().printOutput(f"{stock}:    {e}")
                continue
        return morningIntradayCandle

    def getMorningOpen(df):
        try:
            open = df["open"][0]
        except KeyError: # pragma: no cover
            open = df["open"][df.index.values[0]]
        index = 0
        while np.isnan(open) and index < len(df):
            try:
                open = df["open"][index + 1]
            except KeyError: # pragma: no cover
                open = df["open"][df.index.values[index + 1]]
            index += 1
        return open
    
    def getMorningClose(df):
        try:
            close = df["close"][-1]
        except KeyError: # pragma: no cover
            close = df["close"][df.index.values[-1]]
        index = len(df)
        while np.isnan(close) and index >= 0:
            try:
                close = df["close"][index - 1]
            except KeyError: # pragma: no cover
                close = df["close"][df.index.values[index - 1]]
            index -= 1
        return close
    
    @Halo(text='  [+] Updating candles...', spinner='dots')
    def combineDailyStockDataWithMorningSimulation(allDailyCandles,morningIntradayCandle):
        mutableAllDailyCandles = copy.deepcopy(allDailyCandles)
        stocks = list(mutableAllDailyCandles.keys())
        intradayStocks = list(morningIntradayCandle.keys())
        priceDict = {}
        listPriceDict = []
        for stock in stocks:
            try:
                priceDict = {}
                if stock in intradayStocks:
                    morningPrice = round(morningIntradayCandle[stock]["data"][0][3],2)
                    closePrice = round(mutableAllDailyCandles[stock]["data"][-1][3],2)
                    priceDict["Stock"] = stock
                    priceDict["Morning"] = morningPrice
                    priceDict["EoD"] = closePrice
                    listPriceDict.append(priceDict)
                    # We basically need to replace today's candle with a single candle that has data from market open to the time
                    # when we are taking as reference point in the morning. This is how it would have looked when running the scan 
                    # in the morning hours.
                    mutableAllDailyCandles[stock]["data"] = mutableAllDailyCandles[stock]["data"][:-1] + [morningIntradayCandle[stock]["data"][0]]
                    mutableAllDailyCandles[stock]["index"] = mutableAllDailyCandles[stock]["index"][:-1] + morningIntradayCandle[stock]["index"]
                else:
                    # We should ideally have all stocks from intraday and eod matching,
                    # but for whatever reason, if we don't have the stock, we should skip those
                    # stocks from analysis
                    del mutableAllDailyCandles[stock]
            except: # pragma: no cover
                del mutableAllDailyCandles[stock]
                if 'PKDevTools_Default_Log_Level' in os.environ.keys():
                    intradayChange = colorText.miniTabulator().tabulate(
                                        pd.DataFrame(listPriceDict),
                                        headers="keys",
                                        tablefmt=colorText.No_Pad_GridFormat,
                                        showindex=False
                                    ).encode("utf-8").decode(STD_ENCODING)
                    default_logger().debug(intradayChange)
                continue
        return mutableAllDailyCandles

    def runScanForStocksFromMorningTrade(stockListFromMorningTrade,dailyCandleData):
        latest_daily_df = None
        return latest_daily_df

    def diffMorningCandleDataWithLatestDailyCandleData(screen_df,save_df, updatedCandleData, allDailyCandles,runOptionName=None,filteredListOfStocks=[]):
        save_df.reset_index(inplace=True)
        screen_df.reset_index(inplace=True)
        save_df.drop(f"index", axis=1, inplace=True, errors="ignore")
        screen_df.drop(f"index", axis=1, inplace=True, errors="ignore")
        stocks = save_df["Stock"]
        filterStocks = []
        for stock in stocks:
            if stock in filteredListOfStocks:
                filterStocks.append(stock)
        stocks = filterStocks
        save_df = save_df[save_df['Stock'].isin(filteredListOfStocks)]
        df_screenResults = None
        for stk in filteredListOfStocks:
            df_screenResults_filter = screen_df[screen_df['Stock'].astype(str).str.contains(f"NSE%3A{stk}") == True]
            df_screenResults = pd.concat([df_screenResults, df_screenResults_filter], axis=0)
        screen_df = df_screenResults
        eodLTPs = []
        dayHighLTPs = []
        morningTimestamps = []
        morningAlertLTPs = []
        sellTimestamps = []
        dayHighTimestamps = []
        sellLTPs = []
        eodDiffs = []
        dayHighDiffs = []
        sqrOffDiffs = []
        index = 0
        ts = None
        row = None
        scrStats = ScreeningStatistics(PKMarketOpenCloseAnalyser.configManager, default_logger())
        tradingDate = PKDateUtilities.tradingDate()
        DEFAULT_ALERT_TIME = PKDateUtilities.currentDateTime().replace(year=tradingDate.year,month=tradingDate.month,day=tradingDate.day,hour=MarketHours().openHour,minute=MarketHours().openMinute+configManager.morninganalysiscandlenumber)
        morningAlertTime = DEFAULT_ALERT_TIME
        for stock in stocks:
            try:
                # Open, High, Low, Close, Adj Close, Volume. We need the 3rd index item: Close.
                dayHighLTP = allDailyCandles[stock]["data"][-1][1]
                endOfDayLTP = allDailyCandles[stock]["data"][-1][3]
                try:
                    savedMorningLTP = updatedCandleData[stock]["data"][-1][3]
                    morningTime = PKDateUtilities.utc_to_ist(updatedCandleData[stock]["index"][-1]).strftime("%H:%M")
                    morningAlertTime = updatedCandleData[stock]["index"][-1]
                except: # pragma: no cover
                    savedMorningLTP = round(save_df["LTP"][index],2)
                    morningTime = DEFAULT_ALERT_TIME.strftime("%H:%M")
                    morningAlertTime = DEFAULT_ALERT_TIME
                morningLTP = savedMorningLTP if pd.notna(savedMorningLTP) else round(save_df["LTP"][index],2)
                morningTimestamps.append(morningTime)
                morningCandles = PKMarketOpenCloseAnalyser.allIntradayCandles
                df = pd.DataFrame(data=morningCandles[stock]["data"],
                                columns=morningCandles[stock]["columns"],
                                index=morningCandles[stock]["index"])
                # try:
                #     # Let's only consider those candles that are after the alert issue-time in the mornings
                #     df = df[df.index >=  pd.to_datetime(f'{PKDateUtilities.tradingDate().strftime(f"%Y-%m-%d")} 09:{15+PKMarketOpenCloseAnalyser.configManager.morninganalysiscandlenumber}:00+05:30').to_datetime64()]
                # except: # pragma: no cover
                #     df = df[df.index >=  pd.to_datetime(f'{PKDateUtilities.tradingDate().strftime(f"%Y-%m-%d")} 09:{15+PKMarketOpenCloseAnalyser.configManager.morninganalysiscandlenumber}:00+05:30', utc=True)]
                #     pass
                ts, row = scrStats.findMACDCrossover(df=df,
                                           afterTimestamp=morningAlertTime,
                                           nthCrossover=1,
                                           upDirection=True)
                # saveDictionary = {}
                # screeningDictionary = {}
                # nextSellMinute = 1
                # foundNextSellCandle = False
                # index = None
                # while not foundNextSellCandle:
                #     try:
                #         # Let's only consider those candles that are right after the alert issue-time in the mornings
                #         index = pd.to_datetime(f'{PKDateUtilities.tradingDate().strftime(f"%Y-%m-%d")} {MarketHours().openHour:02}:{MarketHours().openMinute+PKMarketOpenCloseAnalyser.configManager.morninganalysiscandlenumber+nextSellMinute}:00+05:30').to_datetime64()
                #         df = df[df.index <=  index]
                #     except: # pragma: no cover
                #         index = pd.to_datetime(f'{PKDateUtilities.tradingDate().strftime(f"%Y-%m-%d")} {MarketHours().openHour:02}:{MarketHours().openMinute+PKMarketOpenCloseAnalyser.configManager.morninganalysiscandlenumber+nextSellMinute}:00+05:30', utc=True)
                #         df = df[df.index <=  index]
                #         pass
                #     foundNextSellCandle = scrStats.findATRTrailingStops(df=df,sensitivity=configManager.atrTrailingStopSensitivity, atr_period=configManager.atrTrailingStopPeriod,ema_period=configManager.atrTrailingStopEMAPeriod,buySellAll=2,saveDict=saveDictionary,screenDict=screeningDictionary)
                #     nextSellMinute += 1
                # if foundNextSellCandle:
                #     ts = df.tail(len(df)-index +1).head(1).index[-1]
                #     row = df[df.index == ts]
                highTS, highRow = scrStats.findIntradayHighCrossover(df=df)
                # buySell_df = scrStats.computeBuySellSignals(updatedCandleData[stock]["data"])
                # OutputControls().printOutput(buySell_df)
                dayHighLTP = dayHighLTP if pd.notna(dayHighLTP) else highRow["high"][-1]
                sellTimestamps.append(PKDateUtilities.utc_to_ist(ts).strftime("%H:%M"))
                dayHighTimestamps.append(PKDateUtilities.utc_to_ist(highTS).strftime("%H:%M"))
                sellLTPs.append(row["high"][-1])
                eodLTPs.append(round(endOfDayLTP,2))
                dayHighLTPs.append(round(dayHighLTP,2))
                eodDiffs.append(round(endOfDayLTP - morningLTP,2))
                dayHighDiffs.append(round(dayHighLTP - morningLTP,2))
                sqrOffDiffs.append(round(row["high"][-1] - morningLTP,2))
                morningAlertLTPs.append(str(int(round(morningLTP,0))))
                index += 1
            except: # pragma: no cover
                eodLTPs.append("0")
                eodDiffs.append("0")
                dayHighLTPs.append("0")
                dayHighDiffs.append("0")
                if len(morningAlertLTPs) < len(eodLTPs):
                    morningAlertLTPs.append("0")
                if len(morningTimestamps) < len(eodLTPs):
                    morningTimestamps.append("09:30")
                if len(sellTimestamps) < len(eodLTPs):
                    sellTimestamps.append("09:40")
                if len(sellLTPs) < len(eodLTPs):
                    sellLTPs.append("0")
                if len(sqrOffDiffs) < len(eodLTPs):
                    sqrOffDiffs.append("0")
                if len(dayHighTimestamps) < len(eodLTPs):
                    dayHighTimestamps.append("09:45")
                continue
        diffColumns = ["LTP@Alert", "AlertTime", "SqrOff", "SqrOffLTP", "SqrOffDiff","DayHighTime","DayHigh","DayHighDiff", "EoDLTP", "EoDDiff"]
        diffValues = [morningAlertLTPs, morningTimestamps, sellTimestamps, sellLTPs, sqrOffDiffs,dayHighTimestamps,dayHighLTPs, dayHighDiffs,eodLTPs, eodDiffs]
        for column in diffColumns:
            columnName = column
            save_df[columnName] = diffValues[diffColumns.index(columnName)]
            screen_df.loc[:, columnName] = save_df.loc[:, columnName].apply(
                lambda x: x if columnName in ["LTP@Alert","AlertTime", "SqrOff", "SqrOffLTP", "EoDLTP","DayHigh","DayHighTime"] else ((colorText.GREEN if float(x) >= 0 else colorText.FAIL) + str(x) + colorText.END)
            )

        columns = save_df.columns
        lastIndex = len(save_df)
        ltpSum = 0
        for col in columns:
            if col in ["Stock", "LTP@Alert", "Pattern", "LTP", "SqrOffLTP","SqrOffDiff","DayHigh","DayHighDiff", "EoDLTP", "EoDDiff", "%Chng"]:
                if col == "Stock":
                    save_df.loc[lastIndex,col] = "BASKET"
                elif col == "Pattern":
                    save_df.loc[lastIndex,col] = runOptionName if runOptionName is not None else ""
                elif col in ["LTP", "LTP@Alert", "SqrOffLTP","SqrOffDiff", "EoDLTP", "EoDDiff","DayHigh","DayHighDiff"]:
                    save_df.loc[lastIndex,col] = round(sum(save_df[col].dropna(inplace=False).astype(float)),2)
                elif col == "%Chng":
                    ltpSum = sum(save_df["LTP@Alert"].dropna(inplace=False).astype(float))
                    change_pct = sum(save_df["EoDDiff"].dropna(inplace=False).astype(float))*100/ltpSum
                    save_df.loc[lastIndex,col] = f"{round(change_pct,2)}%"
            else:
                save_df.loc[lastIndex,col] = ""
            screen_df.loc[lastIndex,col] = save_df.loc[lastIndex,col]
        eodDiff = save_df.loc[lastIndex,"EoDDiff"]
        sqrOffDiff = save_df.loc[lastIndex,"SqrOffDiff"]
        dayHighDiff = save_df.loc[lastIndex,"DayHighDiff"]
        save_df.loc[lastIndex,"EoDDiff"] = str(eodDiff) + f'({round(100*eodDiff/ltpSum,2) if ltpSum >0 else 0}%)'
        save_df.loc[lastIndex,"SqrOffDiff"] = str(sqrOffDiff) + f'({round(100*sqrOffDiff/ltpSum,2) if ltpSum >0 else 0}%)'
        save_df.loc[lastIndex,"DayHighDiff"] = str(dayHighDiff) + f'({round(100*dayHighDiff/ltpSum,2) if ltpSum >0 else 0}%)'
        screen_df.loc[lastIndex,"EoDDiff"] = (colorText.GREEN if eodDiff >= 0 else colorText.FAIL) + save_df.loc[lastIndex,"EoDDiff"] + colorText.END
        screen_df.loc[lastIndex,"SqrOffDiff"] = (colorText.GREEN if sqrOffDiff >= 0 else colorText.FAIL) + save_df.loc[lastIndex,"SqrOffDiff"] + colorText.END
        screen_df.loc[lastIndex,"DayHighDiff"] = (colorText.GREEN if dayHighDiff >= 0 else colorText.FAIL) + save_df.loc[lastIndex,"DayHighDiff"] + colorText.END
        save_df.set_index("Stock", inplace=True)
        screen_df.set_index("Stock", inplace=True)
        PKMarketOpenCloseAnalyser.allIntradayCandles = None
        screen_df.replace(np.nan, "", regex=True)
        save_df.replace(np.nan, "", regex=True)
        # Drop the unnecessary columns for this scanner type to make way for other columns to be fitted nicely on screen
        columnsToBeDropped = ["Breakout(22Prds)","index","EoDLTP","RS_Rating^NSEI","RVM(15)"]
        for col in columnsToBeDropped:
            if col in save_df.columns:
                save_df.drop(col, axis=1, inplace=True, errors="ignore")
            if col in screen_df.columns:
                screen_df.drop(col, axis=1, inplace=True, errors="ignore")
        return save_df, screen_df



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""

from PKDevTools.classes.OutputControls import OutputControls
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.log import default_logger

from pkscreener.classes import ConfigManager
from pkscreener.classes.ConsoleUtility import PKConsoleTools
from pkscreener.classes.MenuOptions import menus

# Class for managing misc console menu utility methods
class PKConsoleMenuTools:
    configManager = ConfigManager.tools()
    configManager.getConfig(ConfigManager.parser)

    # Prompt for asking RSI
    def promptRSIValues():
        PKConsoleTools.clearScreen(forceTop=True)
        try:
            minRSI, maxRSI = int(
                input(
                    colorText.WARN
                    + "\n  [+] Enter Min RSI value (Default=55): "
                    + colorText.END
                ) or 55
            ), int(
                input(
                    colorText.WARN
                    + "  [+] Enter Max RSI value (Default=68): "
                    + colorText.END
                ) or "68"
            )
            if (
                (minRSI >= 0 and minRSI <= 100)
                and (maxRSI >= 0 and maxRSI <= 100)
                and (minRSI <= maxRSI)
            ):
                return (minRSI, maxRSI)
            raise ValueError
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            return (0, 0)

    # Prompt for asking CCI
    def promptCCIValues(minCCI=None, maxCCI=None):
        PKConsoleTools.clearScreen(forceTop=True)
        if minCCI is not None and maxCCI is not None:
            return minCCI, maxCCI
        try:
            minCCI, maxCCI = int(
                input(
                    colorText.WARN
                    + "\n  [+] Enter Min CCI value (Default=110): "
                    + colorText.END
                ) or "110"
            ), int(
                input(
                    colorText.WARN
                    + "  [+] Enter Max CCI value (Default=300): "
                    + colorText.END
                ) or "300"
            )
            if minCCI <= maxCCI:
                return (minCCI, maxCCI)
            raise ValueError
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            return (-100, 100)

    # Prompt for asking Volume ratio
    def promptVolumeMultiplier(volumeRatio=None):
        PKConsoleTools.clearScreen(forceTop=True)
        if volumeRatio is not None:
            return volumeRatio
        try:
            volumeRatio = float(
                input(
                    colorText.WARN
                    + "\n  [+] Enter Min Volume ratio value (Default = 2.5): "
                    + colorText.END
                ) or "2.5"
            )
            if volumeRatio > 0:
                return volumeRatio
            raise ValueError
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            return 2

    def promptMenus(menu):
        PKConsoleTools.clearScreen(forceTop=True)
        m = menus()
        m.level = menu.level if menu is not None else 0
        return m.renderForMenu(menu)

    def promptChartPatternSubMenu(menu,respChartPattern):
        PKConsoleTools.clearScreen(forceTop=True)
        m3 = menus()
        m3.renderForMenu(menu,asList=True)
        lMenu =  m3.find(str(respChartPattern))
        maLength = PKConsoleMenuTools.promptSubMenuOptions(lMenu,defaultOption= "4" if respChartPattern == 3 else "1" )
        return maLength
    
    # Prompt for submenu options
    def promptSubMenuOptions(menu=None, defaultOption="1"):
        try:
            PKConsoleMenuTools.promptMenus(menu=menu)
            resp = int(
                input(
                    colorText.WARN
                    + """  [+] Select Option:"""
                    + colorText.END
                ) or defaultOption
            )
            if resp >= 0 and resp <= 10:
                return resp
            raise ValueError
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().takeUserInput(
                colorText.FAIL
                + "\n  [+] Invalid Option Selected. Press <Enter> to try again..."
                + colorText.END
            )
            return None

    # Prompt for Reversal screening
    def promptReversalScreening(menu=None):
        try:
            PKConsoleMenuTools.promptMenus(menu=menu)
            resp = int(
                input(
                    colorText.WARN
                    + """  [+] Select Option:"""
                    + colorText.END
                ) or "3"
            )
            if resp >= 0 and resp <= 10:
                if resp == 4:
                    try:
                        defaultMALength = 9 if PKConsoleMenuTools.configManager.duration.endswith("m") else 50
                        maLength = int(
                            input(
                                colorText.WARN
                                + f"\n  [+] Enter MA Length (E.g. 9,10,20,50 or 200) (Default={defaultMALength}): "
                                + colorText.END
                            ) or str(defaultMALength)
                        )
                        return resp, maLength
                    except ValueError as e:  # pragma: no cover
                        default_logger().debug(e, exc_info=True)
                        OutputControls().printOutput(
                            colorText.FAIL
                            + "\n[!] Invalid Input! MA Length should be single integer value!\n"
                            + colorText.END
                        )
                        raise ValueError
                elif resp == 6:
                    try:
                        maLength = int(
                            input(
                                colorText.WARN
                                + "\n  [+] Enter NR timeframe [Integer Number] (E.g. 4, 7, etc.) (Default=4): "
                                + colorText.END
                            ) or "4"
                        )
                        return resp, maLength
                    except ValueError as e:  # pragma: no cover
                        default_logger().debug(e, exc_info=True)
                        OutputControls().printOutput(
                            colorText.FAIL
                            + "\n[!] Invalid Input! NR timeframe should be single integer value!\n"
                            + colorText.END
                        )
                        raise ValueError
                elif resp in [7,10]:
                    m3 = menus()
                    m3.renderForMenu(menu,asList=True)
                    lMenu =  m3.find(str(resp))
                    return resp, PKConsoleMenuTools.promptSubMenuOptions(lMenu)
                return resp, None
            raise ValueError
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().takeUserInput(
                colorText.FAIL
                + "\n  [+] Invalid Option Selected. Press <Enter> to try again..."
                + colorText.END
            )
            return None, None

    # Prompt for Reversal screening
    def promptChartPatterns(menu=None):
        try:
            PKConsoleMenuTools.promptMenus(menu=menu)
            resp = int(
                input(
                    colorText.WARN
                    + """  [+] Select Option:"""
                    + colorText.END
                ) or "3"
            )
            if resp == 1 or resp == 2:
                candles = int(
                    input(
                        colorText.WARN
                        + "\n  [+] How many candles (TimeFrame) to look back Inside Bar formation? (Default=3): "
                        + colorText.END
                    ) or "3"
                )
                return (resp, candles)
            if resp == 3:
                percent = float(
                    input(
                        colorText.WARN
                        + "\n  [+] Enter Percentage within which all MA/EMAs should be (Ideal: 0.1-2%)? (Default=0.8): "
                        + colorText.END
                    ) or "0.8"
                )
                return (resp, percent / 100.0)
            if resp >= 0 and resp <= 9:
                return resp, 0
            raise ValueError
        except ValueError as e:  # pragma: no cover
            default_logger().debug(e, exc_info=True)
            OutputControls().takeUserInput(
                colorText.FAIL
                + "\n  [+] Invalid Option Selected. Press <Enter> to try again..."
                + colorText.END
            )
            return (None, None)



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import warnings

warnings.simplefilter("ignore", DeprecationWarning)
warnings.simplefilter("ignore", FutureWarning)
import pandas as pd
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.PKDateUtilities import PKDateUtilities
from PKDevTools.classes.log import default_logger
from pkscreener.classes import Utility, ConsoleUtility
from pkscreener.classes.ConfigManager import parser, tools

configManager = tools()
configManager.getConfig(parser)

# Backtests for a given stock with the data for the past x number of days.
# Before this gets called, the assumption is that the user must already
# have run some scanner from the available options. SampleDays is the
# sampling period for which we need to run the backtests. Generally, a
# 30 day period or a 180-day period should be more than enough.
def backtest(
    stock,
    data,
    saveDict=None,
    screenedDict=None,
    periods=30,
    sampleDays=configManager.backtestPeriod,
    backTestedData=None,
    sellSignal=False,
):
    if stock == "" or data is None:
        default_logger().debug(f"No data/stock {(stock)} received for backtesting!")
        return
    if screenedDict is None or len(screenedDict) == 0:
        default_logger().debug(f"{(stock)}No backtesting strategy or screened dictionary received!")
        return
    calcPeriods = configManager.periodsRange
    allStockBacktestData = []
    # Take the data based on which the result set for a strategy may have been arrived at
    # The results must have been arrived at with data based on configManager.backtestPeriod -sampleDays
    # but we also need the periods days to be able to calculate the next few days' returns
    # s1    d0
    # s1    d1
    # s1    d2  <----------------On this day the recommendation was made
    # s1    d3  ^
    #   ....    |
    # s1    dn  |----------------We need to make calculations upto 30 day period from d2
    previous_recent = data.head(
        1
    )  # This is the row which has the date for which the recommendation is valid
    if len(previous_recent) <= 0:
        return backTestedData
    data = data.head(max(calcPeriods) + 1)
    # Let's check the returns for the given strategy over a period ranging from 1 period to 30 periods.
    # columns=['Stock', 'Date', "volume", 'Trend', 'MA-Signal', 'LTP', '52Wk-H',
    #          '52Wk-L', '1-Pd', '2-Pd', '3-Pd', '4-Pd', '5-Pd', '10-Pd', '15-Pd',
    #          '22-Pd', '30-Pd', 'Consol.', 'Breakout', 'RSI', 'Pattern', 'CCI',
    #          'LTP1', 'Growth1', 'LTP2', 'Growth2', 'LTP3', 'Growth3', 'LTP4',
    #          'Growth4', 'LTP5', 'Growth5', 'LTP10', 'Growth10', 'LTP15', 'Growth15',
    #          'LTP22', 'Growth22', 'LTP30', 'Growth30']
    
    # incoming = list(saveDict.keys())
    # for prd in calcPeriods:
    #     colNames = [f"LTP{prd}",f"Growth{prd}",f"{prd}-Pd"]
    #     if colNames[0] not in incoming:
    #         for col in colNames:
    #             columns.remove(col)
    # if backTestedData is None:
    #     backTestedData = pd.DataFrame(columns=columns)
    # df = pd.DataFrame([screenedDict],columns=columns)
    # df[f"LTP{prd}"] = saveDict[f"LTP{prd}"]
    # df[f"Growth{prd}"] = saveDict[f"Growth{prd}"]
    columns=[
                "Stock",
                "Date",
                "volume",
                "Trend",
                "MA-Signal",
                "LTP",
                "52Wk-H",
                "52Wk-L"
            ]
    backTestedStock = {
        "Stock": "",
        "Date": "",
        "volume": "",
        "Trend": "",
        "MA-Signal": "",
        "LTP": "",
        "52Wk-H": "",
        "52Wk-L": ""
    }
    for prd in calcPeriods:
        columns.append(f"{prd}-Pd")
        backTestedStock[f"{prd}-Pd"] = ""
    if backTestedData is None:
        backTestedData = pd.DataFrame(columns=columns)
    backTestedStock["Stock"] = stock
    backTestedStock["Date"] = saveDict["Date"]
    backTestedStock["Consol."] = screenedDict["Consol."]
    backTestedStock["Breakout"] = screenedDict["Breakout"]
    backTestedStock["MA-Signal"] = screenedDict["MA-Signal"]
    backTestedStock["volume"] = screenedDict["volume"]
    backTestedStock["LTP"] = screenedDict["LTP"]
    backTestedStock["52Wk-H"] = screenedDict["52Wk-H"]
    backTestedStock["52Wk-L"] = screenedDict["52Wk-L"]
    backTestedStock["RSI"] = screenedDict["RSI"]
    backTestedStock["Trend"] = screenedDict["Trend"]
    backTestedStock["Pattern"] = screenedDict["Pattern"]
    backTestedStock["CCI"] = screenedDict["CCI"]
    for prd in calcPeriods:
        try:
            backTestedStock[f"{abs(prd)}-Pd"] = ""
            backTestedStock[f"LTP{prd}"] = ""
            backTestedStock[f"Growth{prd}"] = ""
            rolling_pct = data["close"].pct_change(periods=prd) * 100
            pct_change = rolling_pct.iloc[prd]
            if not sellSignal:
                colored_pct = colorText.GREEN if pct_change >= 0 else colorText.FAIL
            else:
                colored_pct = colorText.FAIL if pct_change >= 0 else colorText.GREEN
            backTestedStock[f"{abs(prd)}-Pd"] = (
                colored_pct + "%.2f%%" % pct_change + colorText.END
            )
        except Exception:# pragma: no cover
            pass
        # Let's capture the portfolio data, if available
        try:
            backTestedStock[f"LTP{prd}"] = saveDict[f"LTP{prd}"]
            backTestedStock[f"Growth{prd}"] = saveDict[f"Growth{prd}"]
        except Exception:# pragma: no cover
            pass
        # else:
        #     del backTestedStock[f"{abs(prd)}-Pd"]
        #     try:
        #         backTestedData = backTestedData.drop(f"{abs(prd)}-Pd", axis=1)
        #     except Exception:
        #         continue
    allStockBacktestData.append(backTestedStock)
    df = pd.DataFrame(allStockBacktestData)  # , columns=backTestedData.columns)
    try:
        backTestedData = pd.concat([backTestedData, df])
    except Exception:# pragma: no cover
        pass
    return backTestedData

# Prepares a backtest summary based on the color codes of individual days or stocks
# Based on that it calculates an overall success rate of a given strategy for which
# this backtest is run.
def backtestSummary(df):
    summary = {}
    overall = {}
    summaryList = []
    net_positives = 0
    net_negatives = 0
    if df is None:
        return
    df.drop_duplicates(inplace=True)
    df_grouped = df.groupby("Stock")
    for col in df.keys():
        if str(col).endswith("-Pd"):
            overall[col] = [0, 0]
    # iterate over each group of stock rows
    for stock_name, df_group in df_grouped:
        # for row_index, row in df_group.iterrows():
        group_positives = 0
        group_negatives = 0
        summary["Stock"] = stock_name
        for col in df_group.keys():
            if str(col).endswith("-Pd"):
                col_positives = (
                    df_group[col]
                    .astype(str)
                    .str.count(colorText.GREEN.replace("[", "\["))
                    .sum()
                )
                col_negatives = (
                    df_group[col]
                    .astype(str)
                    .str.count(colorText.FAIL.replace("[", "\["))
                    .sum()
                )
                group_positives += col_positives
                group_negatives += col_negatives
                overall[col] = [
                    overall[col][0] + col_positives,
                    overall[col][1] + col_negatives,
                ]
                overAllPeriodPrediction = (
                    col_positives * 100 / (col_positives + col_negatives)
                )
                if col_positives + col_negatives == 0:
                    summary[col] = "-"
                else:
                    summary[
                        col
                    ] = f"{ConsoleUtility.PKConsoleTools.formattedBacktestOutput(overAllPeriodPrediction)} of ({col_positives+col_negatives})"
        overAllRowPrediction = (
            group_positives * 100 / (group_positives + group_negatives)
        )
        if group_positives + group_negatives == 0:
            summary["Overall"] = "-"
        else:
            summary[
                "Overall"
            ] = f"{ConsoleUtility.PKConsoleTools.formattedBacktestOutput(overAllRowPrediction)} of ({group_positives+group_negatives})"
        summaryList.append(summary)
        summary = {}
        net_positives += group_positives
        net_negatives += group_negatives

    # Now prepare overall summary
    summary["Stock"] = "SUMMARY"
    for col in overall.keys():
        col_positives = overall[col][0]
        col_negatives = overall[col][1]
        if col_positives + col_negatives == 0:
            summary[col] = "-"
        else:
            summary[
                col
            ] = f"{ConsoleUtility.PKConsoleTools.formattedBacktestOutput((col_positives*100/(col_positives+col_negatives)))} of ({col_positives+col_negatives})"
    if net_positives + net_negatives == 0:
        summary["Overall"] = "-"
    else:
        summary[
            "Overall"
        ] = f"{ConsoleUtility.PKConsoleTools.formattedBacktestOutput(net_positives*100/(net_positives+net_negatives))} of ({net_positives+net_negatives})"
    summaryList.append(summary)
    summary_df = pd.DataFrame(summaryList, columns=summary.keys())
    return summary_df



"""
    The MIT License (MIT)

    Copyright (c) 2023 pkjmesra

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"""
import sys
from PKDevTools.classes.ColorText import colorText
from PKDevTools.classes.OutputControls import OutputControls

import pkscreener.classes.ConfigManager as ConfigManager
import pkscreener.classes.Utility as Utility
from pkscreener.classes import AssetsManager

class UserMenuChoicesHandler:
    configManager = ConfigManager.tools()
    configManager.getConfig(ConfigManager.parser)

    def getDownloadChoices(defaultAnswer=None):
        global userPassedArgs
        argsIntraday = userPassedArgs is not None and userPassedArgs.intraday is not None
        intradayConfig = UserMenuChoicesHandler.configManager.isIntradayConfig()
        intraday = intradayConfig or argsIntraday
        exists, cache_file = AssetsManager.PKAssetsManager.afterMarketStockDataExists(intraday)
        if exists:
            shouldReplace = AssetsManager.PKAssetsManager.promptFileExists(
                cache_file=cache_file, defaultAnswer=defaultAnswer
            )
            if shouldReplace == "N":
                OutputControls().printOutput(
                    cache_file
                    + colorText.END
                    + " already exists. Exiting as user chose not to replace it!"
                )
                sys.exit(0)
            else:
                pattern = f"{'intraday_' if intraday else ''}stock_data_*.pkl"
                UserMenuChoicesHandler.configManager.deleteFileWithPattern(pattern)
        return "X", 12, 0, {"0": "X", "1": "12", "2": "0"}

    def getTopLevelMenuChoices(startupoptions, testBuild, downloadOnly, defaultAnswer=None):
        global selectedChoice, userPassedArgs
        executeOption = None
        menuOption = None
        indexOption = None
        options = []
        if startupoptions is not None:
            options = startupoptions.split(":")
            menuOption = options[0] if len(options) >= 1 else None
            indexOption = options[1] if len(options) >= 2 else None
            executeOption = options[2] if len(options) >= 3 else None
        if testBuild:
            menuOption, indexOption, executeOption, selectedChoice = UserMenuChoicesHandler.getTestBuildChoices(
                indexOption=indexOption,
                executeOption=executeOption,
                menuOption=menuOption,
            )
        elif downloadOnly:
            menuOption, indexOption, executeOption, selectedChoice = UserMenuChoicesHandler.getDownloadChoices(
                defaultAnswer=defaultAnswer
            )
            intraday = userPassedArgs.intraday or UserMenuChoicesHandler.configManager.isIntradayConfig()
            filePrefix = "INTRADAY_" if intraday else ""
            _, cache_file_name = AssetsManager.PKAssetsManager.afterMarketStockDataExists(intraday)
            Utility.tools.set_github_output(f"{filePrefix}DOWNLOAD_CACHE_FILE_NAME",cache_file_name)
        return options, menuOption, indexOption, executeOption
    
    def getTestBuildChoices(indexOption=None, executeOption=None, menuOption=None):
        if menuOption is not None:
            return (
                str(menuOption),
                indexOption if indexOption is not None else 1,
                executeOption if executeOption is not None else 0,
                {
                    "0": str(menuOption),
                    "1": (str(indexOption) if indexOption is not None else 1),
                    "2": (str(executeOption) if executeOption is not None else 0),
                },
            )
        return "X", 1, 0, {"0": "X", "1": "1", "2": "0"}
    
    def handleExitRequest(executeOption):
        if executeOption == "Z":
            input(
                colorText.FAIL
                + "  [+] Press <Enter> to Exit!"
                + colorText.END
            )
            sys.exit(0)
    
    


"""
Main application refactoring for PKScreener
This module breaks down the massive main function into manageable classes
while preserving all existing functionality.
"""

import multiprocessing
import os
import sys
import time
import pandas as pd
from datetime import timedelta
from typing import Dict, List, Tuple, Any, Optional

# Import all necessary modules (preserving original imports)
import urllib.error
from alive_progress import alive_bar

# Import project-specific modules
from pkscreener.classes import colorText, Utility, PKDateUtilities, ConsoleUtility, PKScanRunner
from pkscreener.classes.ConfigManager import ConfigManager
from pkscreener.classes.Fetcher import fetcher
from pkscreener.classes.PKMenu import m0, m1, m2, m3, m4
from pkscreener.classes.PortfolioXRay import PortfolioXRay
from pkscreener.classes.log import default_logger
from pkscreener.classes.PKAnalytics import PKAnalyticsService
from pkscreener.classes.PKPremiumHandler import PKPremiumHandler
from pkscreener.classes.PKSpreadsheets import PKSpreadsheets
from pkscreener.globals import *

class ApplicationState:
    """Manages the global application state and configuration"""
    
    def __init__(self, userArgs=None):
        self.userArgs = userArgs
        self.initialize_globals()
        self.setup_initial_state()
        
    def initialize_globals(self):
        """Initialize all global variables with their default values"""
        global lastScanOutputStockCodes, scanCycleRunning, runCleanUp, test_messages_queue
        global show_saved_diff_results, criteria_dateTime, analysis_dict, mp_manager
        global listStockCodes, screenResults, selectedChoice, defaultAnswer, menuChoiceHierarchy
        global screenCounter, screenResultsCounter, stockDictPrimary, stockDictSecondary
        global userPassedArgs, loadedStockData, keyboardInterruptEvent, loadCount, maLength
        global newlyListedOnly, keyboardInterruptEventFired, strategyFilter, elapsed_time, start_time
        
        selectedChoice = {"0": "", "1": "", "2": "", "3": "", "4": ""}
        elapsed_time = 0 if not scanCycleRunning else elapsed_time
        start_time = 0 if not scanCycleRunning else start_time
        
    def setup_initial_state(self):
        """Set up the initial application state based on user arguments"""
        global testing, testBuild, downloadOnly, startupoptions, user, defaultAnswer, userPassedArgs
        global runOptionName, options, strategyFilter, test_messages_queue
        
        testing = False if self.userArgs is None else (self.userArgs.testbuild and self.userArgs.prodbuild)
        testBuild = False if self.userArgs is None else (self.userArgs.testbuild and not testing)
        downloadOnly = False if self.userArgs is None else self.userArgs.download
        startupoptions = None if self.userArgs is None else self.userArgs.options
        user = None if self.userArgs is None else self.userArgs.user
        defaultAnswer = None if self.userArgs is None else self.userArgs.answerdefault
        userPassedArgs = self.userArgs
        runOptionName = ""
        options = []
        strategyFilter = []
        test_messages_queue = []
        
        describeUser()


class MenuHandler:
    """Handles all menu-related operations and user interactions"""
    
    def __init__(self, app_state):
        self.app_state = app_state
        self.configManager = ConfigManager()
        
    def process_top_level_menu(self):
        """Process the top-level menu options"""
        global screenResults, saveResults, options, menuOption, indexOption, executeOption
        
        screenResults, saveResults = PKScanRunner.initDataframes()
        options, menuOption, indexOption, executeOption = getTopLevelMenuChoices(
            self.app_state.startupoptions, self.app_state.testBuild, 
            self.app_state.downloadOnly, defaultAnswer=self.app_state.defaultAnswer
        )
        
        selectedMenu = initExecution(menuOption=menuOption)
        menuOption = selectedMenu.menuKey
        
        return selectedMenu, menuOption, indexOption, executeOption
    
    def handle_premium_checks(self, selectedMenu):
        """Check if user has premium access for premium features"""
        global menuOption, indexOption, executeOption
        
        if menuOption in ["F", "M", "S", "B", "G", "C", "P", "D"] or selectedMenu.isPremium:
            ensureMenusLoaded(menuOption, indexOption, executeOption)
            if not PKPremiumHandler.hasPremium(selectedMenu):
                PKAnalyticsService().send_event(f"non_premium_user_{menuOption}_{indexOption}_{executeOption}")
                PKAnalyticsService().send_event("app_exit")
                sys.exit(0)
    
    def handle_special_menu_options(self, menuOption, selectedMenu):
        """Handle special menu options like monitor, download, etc."""
        if menuOption in ["M", "D", "I", "L", "F"]:
            self.handle_monitor_download_options(menuOption, selectedMenu)
        elif menuOption in ["P"]:
            self.handle_predefined_scans()
        elif menuOption in ["X", "T", "E", "Y", "U", "H", "C"]:
            self.handle_scanner_menu_options()
        elif menuOption in ["B", "G"]:
            self.handle_backtest_options()
        elif menuOption in ["S"]:
            self.handle_strategy_options()
    
    def handle_monitor_download_options(self, menuOption, selectedMenu):
        """Handle monitor and download menu options"""
        launcher = f'"{sys.argv[0]}"' if " " in sys.argv[0] else sys.argv[0]
        launcher = f"python3.12 {launcher}" if (launcher.endswith(".py\"") or launcher.endswith(".py")) else launcher
        
        if menuOption in ["M"]:
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener in monitoring mode. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} --systemlaunched -a Y -m 'X'{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit monitoring mode.{colorText.END}")
            PKAnalyticsService().send_event(f"monitor_{menuOption}")
            time.sleep(2)
            os.system(f"{launcher} --systemlaunched -a Y -m 'X'")
        elif menuOption in ["D"]:
            self.handle_download_options(selectedMenu, launcher)
        elif menuOption in ["L"]:
            PKAnalyticsService().send_event(f"{menuOption}")
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener to collect logs. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} -a Y -l{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit at any time.{colorText.END}")
            time.sleep(2)
            os.system(f"{launcher} -a Y -l")
        elif menuOption in ["F"]:
            self.handle_favorites_option()
    
    def handle_download_options(self, selectedMenu, launcher):
        """Handle download-specific menu options"""
        ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
        m1.renderForMenu(selectedMenu)
        selDownloadOption = input(colorText.FAIL + "  [+] Select option: ") or "D"
        OutputControls().printOutput(colorText.END, end="")
        
        if selDownloadOption.upper() == "D":
            OutputControls().printOutput(f"{colorText.GREEN}Launching PKScreener to Download daily OHLC data. If it does not launch, please try with the following:{colorText.END}\n{colorText.FAIL}{launcher} -a Y -e -d{colorText.END}\n{colorText.WARN}Press Ctrl + C to exit at any time.{colorText.END}")
            PKAnalyticsService().send_event(f"{menuOption}_{selDownloadOption.upper()}")
            time.sleep(2)
            os.system(f"{launcher} -a Y -e -d")
            return None, None
        # ... handle other download options
    
    def handle_favorites_option(self):
        """Handle favorites menu option"""
        global indexOption, selectedChoice, listStockCodes
        
        PKAnalyticsService().send_event(f"{menuOption}")
        indexOption = 0
        selectedChoice["0"] = "F"
        selectedChoice["1"] = "0"
        executeOption = None
        shouldSuppress = not OutputControls().enableMultipleLineOutput
        
        if userPassedArgs is not None and userPassedArgs.options is not None and len(userPassedArgs.options.split(":")) >= 3:
            stockOptions = userPassedArgs.options.split(":")
            stockOptions = userPassedArgs.options.split(":")[2 if len(stockOptions)<=3 else 3]
            listStockCodes = stockOptions.replace(".",",").split(",")
        
        if listStockCodes is None or len(listStockCodes) == 0:
            with SuppressOutput(suppress_stderr=shouldSuppress, suppress_stdout=shouldSuppress):
                listStockCodes = fetcher.fetchStockCodes(tickerOption=0, stockCode=None)
        
        ConsoleUtility.PKConsoleTools.clearScreen(clearAlways=True, forceTop=True)
    
    def handle_predefined_scans(self):
        """Handle predefined scan options"""
        global predefinedOption, selPredefinedOption, selIndexOption, selectedChoice
        
        predefinedOption = None
        selPredefinedOption = None
        selIndexOption = None
        
        if len(options) >= 3:
            predefinedOption = str(options[1]) if str(options[1]).isnumeric() else '1'
            selPredefinedOption = str(options[2]) if str(options[2]).isnumeric() else '1'
            if len(options) >= 4:
                selIndexOption = str(options[3]) if str(options[3]).isnumeric() else '12'
        
        selectedChoice["0"] = "P"
        updateMenuChoiceHierarchy()
        selectedMenu = m0.find(menuOption)
        m1.renderForMenu(selectedMenu, asList=(userPassedArgs is not None and userPassedArgs.options is not None))
        
        # ... handle predefined scan logic


class ScannerHandler:
    """Handles scanner operations and screening logic"""
    
    def __init__(self, app_state):
        self.app_state = app_state
        self.configManager = ConfigManager()
    
    def execute_scanner(self, menuOption, indexOption, executeOption):
        """Execute the selected scanner with appropriate parameters"""
        global volumeRatio, minRSI, maxRSI, insideBarToLookback, respChartPattern
        global daysForLowestVolume, reversalOption, maLength
        
        volumeRatio = self.configManager.volumeRatio
        
        if executeOption == 3:
            userPassedArgs.maxdisplayresults = max(self.configManager.maxdisplayresults, 2000)
        elif executeOption == 4:
            daysForLowestVolume = self.handle_scanner_execute_option_4(executeOption, options)
        elif executeOption == 5:
            self.handle_rsi_scanner(executeOption, options)
        elif executeOption == 6:
            self.handle_reversal_scanner(executeOption, options)
        elif executeOption == 7:
            self.handle_chart_pattern_scanner(executeOption, options)
        # ... handle other execute options
    
    def handle_rsi_scanner(self, executeOption, options):
        """Handle RSI scanner configuration"""
        global minRSI, maxRSI, selectedMenu
        
        selectedMenu = m2.find(str(executeOption))
        if len(options) >= 5:
            if str(options[3]).isnumeric():
                minRSI = int(options[3])
                maxRSI = int(options[4])
            elif str(options[3]).upper() == "D" or userPassedArgs.systemlaunched:
                minRSI = 60
                maxRSI = 75
        else:
            minRSI, maxRSI = ConsoleMenuUtility.PKConsoleMenuTools.promptRSIValues()
        
        if not minRSI and not maxRSI:
            OutputControls().printOutput(
                colorText.FAIL
                + "\n  [+] Error: Invalid values for RSI! Values should be in range of 0 to 100. Please try again!"
                + colorText.END
            )
            OutputControls().takeUserInput("Press <Enter> to continue...")
            return None, None
    
    def handle_reversal_scanner(self, executeOption, options):
        """Handle reversal scanner configuration"""
        global reversalOption, maLength, selectedMenu
        
        selectedMenu = m2.find(str(executeOption))
        if len(options) >= 4:
            reversalOption = int(options[3])
            if reversalOption in [4, 6, 7, 10]:
                if len(options) >= 5:
                    if str(options[4]).isnumeric():
                        maLength = int(options[4])
                    elif str(options[4]).upper() == "D" or userPassedArgs.systemlaunched:
                        maLength = 50 if reversalOption == 4 else (3 if reversalOption in [7] else (2 if reversalOption in [10] else 7))
                elif defaultAnswer == "Y" and user is not None:
                    maLength = 50 if reversalOption == 4 else (3 if reversalOption == 7 else 7)
                else:
                    reversalOption, maLength = ConsoleMenuUtility.PKConsoleMenuTools.promptReversalScreening(selectedMenu)
        else:
            reversalOption, maLength = ConsoleMenuUtility.PKConsoleMenuTools.promptReversalScreening(selectedMenu)
        
        if reversalOption is None or reversalOption == 0 or maLength == 0:
            return None, None
        else:
            selectedChoice["3"] = str(reversalOption)
            if str(reversalOption) in ["7", "10"]:
                selectedChoice["4"] = str(maLength)


class BacktestHandler:
    """Handles backtesting operations and results processing"""
    
    def __init__(self, app_state):
        self.app_state = app_state
        self.configManager = ConfigManager()
    
    def handle_backtest_options(self):
        """Handle backtest menu options"""
        global backtestPeriod
        
        backtestPeriod = 0
        if len(options) >= 2:
            if str(indexOption).isnumeric():
                backtestPeriod = int(indexOption)
            if len(options) >= 4:
                indexOption = executeOption
                executeOption = options[3]
            del options[1]
        
        indexOption, executeOption, backtestPeriod = takeBacktestInputs(
            str(menuOption).upper(), indexOption, executeOption, backtestPeriod
        )
        backtestPeriod = backtestPeriod * self.configManager.backtestPeriodFactor
    
    def process_backtest_results(self, backtest_df):
        """Process and display backtest results"""
        if backtest_df is not None and len(backtest_df) > 0:
            ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
            df_xray = prepareGroupedXRay(backtestPeriod, backtest_df)
            summary_df, sorting, sortKeys = FinishBacktestDataCleanup(backtest_df, df_xray)
            
            while sorting:
                sorting = showSortedBacktestData(backtest_df, summary_df, sortKeys)
            
            if defaultAnswer is None:
                OutputControls().takeUserInput("Press <Enter> to continue...")
        else:
            OutputControls().printOutput("Finished backtesting with no results to show!")


class StrategyHandler:
    """Handles strategy-related operations and filtering"""
    
    def __init__(self, app_state):
        self.app_state = app_state
    
    def handle_strategy_options(self):
        """Handle strategy menu options"""
        global userOption, strategyFilter, menuOption, indexOption, executeOption, selectedChoice
        
        if len(options) >= 2:
            userOption = options[1]
        
        if defaultAnswer is None:
            selectedMenu = m0.find(menuOption)
            m1.strategyNames = PortfolioXRay.strategyNames()
            m1.renderForMenu(selectedMenu=selectedMenu)
            
            try:
                userOption = input(colorText.FAIL + "  [+] Select option: ")
                OutputControls().printOutput(colorText.END, end="")
                
                if userOption == "":
                    userOption = "37"  # NoFilter
                elif userOption == "38":
                    userOption = input(colorText.FAIL + "  [+] Enter Exact Pattern name:")
                    OutputControls().printOutput(colorText.END, end="")
                    
                    if userOption == "":
                        userOption = "37"  # NoFilter
                    else:
                        strategyFilter.append(f"[P]{userOption}")
                        userOption = "38"
            except EOFError:
                userOption = "37"  # NoFilter
            except Exception as e:
                default_logger().debug(e, exc_info=True)
        
        userOption = userOption.upper()
        
        if userOption == "M":
            ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
            return None, None
        elif userOption == "Z":
            handleExitRequest(userOption)
            return None, None
        elif userOption == "S":
            self.handle_strategy_summary()
        else:
            self.apply_strategy_filters(userOption)
    
    def handle_strategy_summary(self):
        """Handle strategy summary display"""
        OutputControls().printOutput(
            colorText.GREEN
            + "  [+] Collecting all metrics for summarising..."
            + colorText.END
        )
        
        savedValue = configManager.showPastStrategyData
        configManager.showPastStrategyData = True
        df_all = PortfolioXRay.summariseAllStrategies()
        
        if df_all is not None and len(df_all) > 0:
            OutputControls().printOutput(
                colorText.miniTabulator().tabulate(
                    df_all,
                    headers="keys",
                    tablefmt=colorText.No_Pad_GridFormat,
                    showindex=False,
                    maxcolwidths=Utility.tools.getMaxColumnWidths(df_all)
                ).encode("utf-8").decode(STD_ENCODING)
            )
            showBacktestResults(df_all, sortKey="Scanner", optionalName="InsightsSummary")
        else:
            OutputControls().printOutput("[!] Nothing to show here yet. Check back later.")
        
        configManager.showPastStrategyData = savedValue
        
        if defaultAnswer is None:
            OutputControls().takeUserInput("Press <Enter> to continue...")
        
        return None, None
    
    def apply_strategy_filters(self, userOption):
        """Apply selected strategy filters"""
        global strategyFilter, menuOption, indexOption, executeOption, selectedChoice
        
        userOptions = userOption.split(",")
        for usrOption in userOptions:
            strategyFilter.append(m1.find(usrOption).menuText.strip())
        
        menuOption, indexOption, executeOption, selectedChoice = getScannerMenuChoices(
            self.app_state.testBuild or self.app_state.testing,
            self.app_state.downloadOnly,
            self.app_state.startupoptions,
            menuOption="X",
            indexOption=indexOption,
            executeOption=executeOption,
            defaultAnswer=self.app_state.defaultAnswer,
            user=self.app_state.user,
        )


class DataLoader:
    """Handles data loading and preparation for screening"""
    
    def __init__(self, app_state):
        self.app_state = app_state
        self.configManager = ConfigManager()
    
    def load_stock_data(self, menuOption, indexOption, downloadOnly, listStockCodes):
        """Load stock data for screening"""
        global stockDictPrimary, stockDictSecondary, loadedStockData
        
        loadedStockData = loadedStockData and stockDictPrimary is not None and len(stockDictPrimary) > 0
        
        if (menuOption in ["X", "B", "G", "S", "F"] and not loadedStockData) or (
            self.configManager.cacheEnabled and not loadedStockData and not self.app_state.testing
        ):
            try:
                import tensorflow as tf
                with tf.device("/device:GPU:0"):
                    stockDictPrimary, stockDictSecondary = loadDatabaseOrFetch(
                        downloadOnly, listStockCodes, menuOption, indexOption
                    )
            except:
                stockDictPrimary, stockDictSecondary = loadDatabaseOrFetch(
                    downloadOnly, listStockCodes, menuOption, indexOption
                )
        
        return len(stockDictPrimary) if stockDictPrimary is not None else 0


class MainApplication:
    """Main application class that orchestrates all components"""
    
    def __init__(self, userArgs=None):
        self.app_state = ApplicationState(userArgs)
        self.menu_handler = MenuHandler(self.app_state)
        self.scanner_handler = ScannerHandler(self.app_state)
        self.backtest_handler = BacktestHandler(self.app_state)
        self.strategy_handler = StrategyHandler(self.app_state)
        self.data_loader = DataLoader(self.app_state)
        self.configManager = ConfigManager()
    
    def main(self, userArgs=None, optionalFinalOutcome_df=None):
        """
        Main entry point for the application
        Refactored from the original massive main function
        
        Args:
            userArgs: Command line arguments passed by user
            optionalFinalOutcome_df: Optional dataframe for final results
            
        Returns:
            Tuple of screen results and save results
        """
        global keyboardInterruptEventFired
        
        # Check for keyboard interrupt
        if keyboardInterruptEventFired:
            return None, None
        
        # Handle intraday analysis if requested
        if self.handle_intraday_analysis(userArgs, optionalFinalOutcome_df):
            savedAnalysisDict = analysis_dict.get(firstScanKey)
            return analysisFinalResults(
                savedAnalysisDict.get("S1"), 
                savedAnalysisDict.get("S2"), 
                optionalFinalOutcome_df, 
                None
            )
        
        # Initialize multiprocessing and market monitor
        self.initialize_multiprocessing()
        self.initialize_market_monitor()
        
        # Handle cleanup if needed
        if self.handle_cleanup(userArgs):
            cleanupLocalResults()
        
        # Process menu options
        selectedMenu, menuOption, indexOption, executeOption = self.menu_handler.process_top_level_menu()
        
        # Check premium access
        self.menu_handler.handle_premium_checks(selectedMenu)
        
        # Handle special menu options
        self.menu_handler.handle_special_menu_options(menuOption, selectedMenu)
        
        # Handle scanner menu options
        if menuOption in ["X", "T", "E", "Y", "U", "H", "C"]:
            menuOption, indexOption, executeOption, selectedChoice = getScannerMenuChoices(
                self.app_state.testBuild or self.app_state.testing,
                self.app_state.downloadOnly,
                self.app_state.startupoptions,
                menuOption=menuOption,
                indexOption=indexOption,
                executeOption=executeOption,
                defaultAnswer=self.app_state.defaultAnswer,
                user=self.app_state.user,
            )
        
        # Handle menu options X, B, G
        handleMenu_XBG(menuOption, indexOption, executeOption)
        
        # Check for exit request
        if str(indexOption).upper() == "M" or str(executeOption).upper() == "M":
            ConsoleUtility.PKConsoleTools.clearScreen(forceTop=True)
            return None, None
        
        # Prepare stocks for screening
        listStockCodes = handleRequestForSpecificStocks(options, indexOption)
        handleExitRequest(executeOption)
        
        if executeOption is None:
            executeOption = 0
        executeOption = int(executeOption)
        
        # Execute the selected scanner
        self.scanner_handler.execute_scanner(menuOption, indexOption, executeOption)
        
        # Load stock data
        loadCount = self.data_loader.load_stock_data(
            menuOption, indexOption, self.app_state.downloadOnly, listStockCodes
        )
        
        # Run the screening process
        screenResults, saveResults = self.run_screening_process(
            menuOption, indexOption, executeOption, listStockCodes, loadCount
        )
        
        # Process results
        if menuOption in ["X", "C", "F"] and (userPassedArgs.monitor is None or self.configManager.alwaysExportToExcel) or ("|" not in userPassedArgs.options and menuOption not in ["B"]):
            finishScreening(
                self.app_state.downloadOnly,
                self.app_state.testing,
                stockDictPrimary,
                self.configManager,
                loadCount,
                self.app_state.testBuild,
                screenResults,
                saveResults,
                self.app_state.user,
            )
        
        # Handle backtest results
        if menuOption == "B":
            self.backtest_handler.process_backtest_results(backtest_df)
        
        # Reset configuration
        resetConfigToDefault()
        
        # Save to Google Sheets if configured
        self.save_to_google_sheets(saveResults)
        
        # Handle pinned menu
        self.handle_pinned_menu(saveResults, menuOption)
        
        return screenResults, saveResults
    
    def run_screening_process(self, menuOption, indexOption, executeOption, listStockCodes, loadCount):
        """Run the screening process with appropriate parameters"""
        global screenResults, saveResults, backtest_df
        
        # Prepare screening parameters
        samplingDuration, fillerPlaceHolder, actualHistoricalDuration = PKScanRunner.getScanDurationParameters(
            self.app_state.testing, menuOption
        )
        
        totalStocksInReview = 0
        savedStocksCount = 0
        downloadedRecently = False
        items = []
        backtest_df = None
        
        # Get progress bar style
        bar, spinner = Utility.tools.getProgressbarStyle()
        
        # Begin screening process
        OutputControls().printOutput(f"{colorText.GREEN}  [+] Adding stocks to the queue...{colorText.END}")
        
        with alive_bar(actualHistoricalDuration, bar=bar, spinner=spinner) as progressbar:
            while actualHistoricalDuration >= 0:
                daysInPast = PKScanRunner.getBacktestDaysForScan(
                    userPassedArgs, backtestPeriod, menuOption, actualHistoricalDuration
                )
                
                try:
                    listStockCodes, savedStocksCount, pastDate = PKScanRunner.getStocksListForScan(
                        userPassedArgs, menuOption, totalStocksInReview, downloadedRecently, daysInPast
                    ) if menuOption not in ["C"] else (listStockCodes, 0, "")
                except KeyboardInterrupt:
                    try:
                        keyboardInterruptEvent.set()
                        keyboardInterruptEventFired = True
                        actualHistoricalDuration = -1
                        break
                    except KeyboardInterrupt:
                        pass
                    OutputControls().printOutput(
                        colorText.FAIL
                        + "\n  [+] Terminating Script, Please wait..."
                        + colorText.END
                    )
                
                exchangeName = "NASDAQ" if (indexOption == 15 or (self.configManager.defaultIndex == 15 and indexOption == 0)) else "INDIA"
                runOptionName = PKScanRunner.getFormattedChoices(userPassedArgs, selectedChoice)
                
                if ((":0:" in runOptionName or "_0_" in runOptionName) and userPassedArgs.progressstatus is not None) or userPassedArgs.progressstatus is not None:
                    runOptionName = userPassedArgs.progressstatus.split("=>")[0].split("  [+] ")[1]
                
                if menuOption in ["F"]:
                    if "^NSEI" in listStockCodes:
                        listStockCodes.remove("^NSEI")
                    items = PKScanRunner.addScansWithDefaultParams(
                        userPassedArgs, self.app_state.testing, self.app_state.testBuild, 
                        newlyListedOnly, self.app_state.downloadOnly, backtestPeriod, 
                        listStockCodes, menuOption, exchangeName, executeOption, volumeRatio, 
                        items, daysInPast, runOption=f"{userPassedArgs.options} =>{runOptionName} => {menuChoiceHierarchy}"
                    )
                else:
                    PKScanRunner.addStocksToItemList(
                        userPassedArgs, self.app_state.testing, self.app_state.testBuild, 
                        newlyListedOnly, self.app_state.downloadOnly, minRSI, maxRSI, 
                        insideBarToLookback, respChartPattern, daysForLowestVolume, 
                        backtestPeriod, reversalOption, maLength, listStockCodes, 
                        menuOption, exchangeName, executeOption, volumeRatio, items, 
                        daysInPast, runOption=f"{userPassedArgs.options} =>{runOptionName} => {menuChoiceHierarchy}"
                    )
                
                if savedStocksCount > 0:
                    progressbar.text(
                        colorText.GREEN
                        + f"Total Stocks: {len(items)}. Added {savedStocksCount} to Stocks from {pastDate} saved from earlier..."
                        + colorText.END
                    )
                
                fillerPlaceHolder = fillerPlaceHolder + 1
                actualHistoricalDuration = samplingDuration - fillerPlaceHolder
                
                if actualHistoricalDuration >= 0:
                    progressbar()
        
        # Run the scan with parameters
        screenResults, saveResults, backtest_df, tasks_queue, results_queue, consumers, logging_queue = PKScanRunner.runScanWithParams(
            userPassedArgs, keyboardInterruptEvent, screenCounter, screenResultsCounter,
            stockDictPrimary, stockDictSecondary, self.app_state.testing, backtestPeriod,
            menuOption, executeOption, samplingDuration, items, screenResults, saveResults,
            backtest_df, scanningCb=runScanners, tasks_queue=tasks_queue, 
            results_queue=results_queue, consumers=consumers, logging_queue=logging_queue
        )
        
        return screenResults, saveResults
    
    def save_to_google_sheets(self, saveResults):
        """Save results to Google Sheets if configured"""
        if 'ALERT_TRIGGER' in os.environ.keys() and os.environ["ALERT_TRIGGER"] == 'Y':
            if "GSHEET_SERVICE_ACCOUNT_DEV" in os.environ.keys() and (userPassedArgs.backtestdaysago is None):
                begin = time.time()
                creds = os.environ.get("GSHEET_SERVICE_ACCOUNT_DEV")
                OutputControls().printOutput(f"{colorText.GREEN}  [+] Saving data to Google Spreadsheets now...{colorText.END}")
                
                gClient = PKSpreadsheets(credentialDictStr=creds)
                runOption = PKScanRunner.getFormattedChoices(userPassedArgs, selectedChoice)
                df = saveResults.copy()
                df["LastTradeDate"], df["LastTradeTime"] = getLatestTradeDateTime(stockDictPrimary)
                
                gClient.df_to_sheet(df=df, sheetName=runOption)
                OutputControls().printOutput(f"{colorText.GREEN} => Done in {round(time.time()-begin,2)}s{colorText.END}")
    
    def handle_pinned_menu(self, saveResults, menuOption):
        """Handle the pinned menu display and options"""
        if ("RUNNER" not in os.environ.keys() and 
            not self.app_state.testing and 
            (userPassedArgs is None or 
             (userPassedArgs is not None and 
              (userPassedArgs.user is None or str(userPassedArgs.user) == DEV_CHANNEL_ID) and 
              (userPassedArgs.answerdefault is None or userPassedArgs.systemlaunched))) and
            not userPassedArgs.testbuild and 
            userPassedArgs.monitor is None):
            
            prevOutput_results = saveResults.index if (saveResults is not None and not saveResults.empty) else []
            isNotPiped = (("|" not in userPassedArgs.options) if (userPassedArgs is not None and userPassedArgs.options is not None) else True)
            hasFoundStocks = len(prevOutput_results) > 0 and isNotPiped
            
            if hasFoundStocks or (self.configManager.showPinnedMenuEvenForNoResult and isNotPiped):
                # ... handle pinned menu logic
                pass


# Replace the original main function with the refactored version
def main(userArgs=None, optionalFinalOutcome_df=None):
    """
    Main entry point - replaces the original massive main function
    
    Args:
        userArgs: Command line arguments passed by user
        optionalFinalOutcome_df: Optional dataframe for final results
        
    Returns:
        Tuple of screen results and save results
    """
    app = MainApplication(userArgs)
    return app.main(userArgs, optionalFinalOutcome_df)
